# -*- coding: utf-8 -*-
from docparser.core.virtual_block import VirtualBlock


class TableVirtualBlock(VirtualBlock):
    """
    表格虚拟区块信息类
    """

    def __init__(self, sheet, block_config):
        VirtualBlock.__init__(self, sheet, block_config)
        self.original_table = OriginalTable(self.rect, block_config)
        self.virtual_table = None

    def _build(self):
        """
        构建虚拟块数据
        """
        return None

    def _build_output(self, virtual_block_data):
        """
        构建输出结果
        """
        return virtual_block_data

    # def _pre_output(self, data):
    #     """
    #     数据前置处理
    #     """
    #     VirtualBlock._pre_output(self, data)
    #
    #     self._to_virtual_table()


class OriginalTable:
    class Header:
        def __init__(self):
            self.rows = []

    class Body:
        def __init__(self):
            self.rows = []

    def __init__(self,rect, table_config):
        self.rect = rect
        self.table_config = table_config
        self.header = OriginalTable.Header()
        self.body = OriginalTable.Body()
        self._build()

    def _build_header(self):
        pass

    def _build_body(self):
        pass

    def _build(self):
        pass


class VirtualTable:
    class Header:
        def __init__(self):
            self.rows = []

    class Body:
        def __init__(self):
            self.rows = []

    class Cell:
        def __init__(self):
            self.Value = None

    def __init__(self, original_table):
        self.original_table = original_table
        self.header = VirtualTable.Header()
        self.body = VirtualTable.Body()



# -*- coding: utf-8 -*-

from marshmallow import Schema, fields

from docparser.config.excel.custom_fields import CustomFields
from docparser.config.excel.rect_config import RectConfigSchema


class BlockConfigSchema(Schema):
    """
    区块配置架构
    """

    # 区块类型
    type = CustomFields.BlockTypeField()

    # 区块数据提取器(standard）
    parser = fields.Str(required=True, default='standard')

    # 区域
    rect = fields.Nested(RectConfigSchema())





# -*- coding: utf-8 -*-
from marshmallow import fields

from docparser.config.excel.block_config import BlockConfigSchema


class TextBlockConfigSchema(BlockConfigSchema):
    """
    文本区块配置架构
    """

    # 正则模式
    pattern = fields.Str()







# -*- coding: utf-8 -*-
import importlib


class VirtualBlockBuilder:
    """
    虚拟区块构建者
    """

    @staticmethod
    def build(sheet, block_config):
        """
        根据类型构建虚拟区块
        """
        try:
            block_type = block_config['type']
            if block_type == 'text':
                parser = block_config['parser'] if 'parser' in block_config['parser'] else 'default'
            else:
                parser = block_config['parser'] if 'parser' in block_config['parser'] else 'standard'

            if parser in ['default', 'mixed', 'standard']:
                module_name = 'common.document.parsers.implements.%s_%s_virtual_block' % (parser, block_type)
            else:
                module_name = 'common.document.parsers.extends.%s_%s_virtual_block' % (parser, block_type)

            class_name = '%sVirtualBlock' % (parser.capitalize(), block_type.capitalize())
            virtual_block_module = importlib.import_module(module_name)
            cls = getattr(virtual_block_module, class_name)

            virtual_block = cls(sheet, block_config)
            virtual_block.build()

            return virtual_block

        except ModuleNotFoundError:
            raise



import json
import traceback
from pymssql import InternalError
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from docparser.core.behavior_base import BehaviorBase


class AddressRepairBehavior(BehaviorBase):
    __sql_filter_arr = ["delete", "insert", "truncate", "drop", "update", "remove", "create", "table", "database"]
    __sql_find_str_filter = ["\"", "\\", "/", "*", "'", "=", "-", "#", ";", "", "+", "%", "$", "(", ")", "@", "!"]
    """
    地址修补处理行为
    =============
    配置说明：主节点下
    "address_repair":{
                    "db":{
                        "pub": {"user":"","pwd":"","server":"","database":""}
                    },
                    "repairs": [
                        {"key":"","table": "","db_key":"","sql":"","column":[],"value":"","mapping":"", "old_val_handle": "empty|retain","try_match":True}
                    ]
                }
    """

    class_index = 2

    def data_processing(self, ref_data, data: list, error: list, config: dict, logger, additional) -> dict:

        sub_conf = config.get("address_repair")
        if data and len(data) > 0 and sub_conf is not None:
            repair_config = sub_conf.get("repairs")
            db_configs = sub_conf.get("db")
            if repair_config is None or db_configs is None:
                return additional

            cache_db_engine = {}
            for conf in repair_config:

                key = conf.get("key")
                db_key = conf.get("db_key")
                table_name = conf.get("table")
                db_conf = db_configs.get(db_key)
                mapping = conf.get("mapping")
                try_match = conf.get("try_match", True)
                old_val_handle = conf.get("old_val_handle", "empty")

                db_engine = self.get_engine(db_key, db_conf, cache_db_engine)

                data_vals = self._find_values(key, table_name, data)
                new_vals = self.get_db_data(db_engine,
                                            data_vals,
                                            conf.get("sql"),
                                            conf.get("column"),
                                            conf.get("value"),
                                            mapping,
                                            try_match
                                            )
                if mapping is not None:
                    if table_name is not None and table_name.strip() != "":
                        data[table_name]["column"].append(mapping)
                        table_index = self._find_col_index(key, table_name, data)
                        for i, v in enumerate(new_vals):
                            if isinstance(v, tuple):
                                data[table_name]["rows"][i].append(v[1])
                                if old_val_handle == "empty":
                                    data[table_name]["rows"][i][table_index] = ""
                    else:
                        if isinstance(new_vals[0], tuple):
                            data[mapping] = new_vals[0][1]
                            if old_val_handle == "empty":
                                data[key] = ""
                else:
                    self._restore_values(key, table_name, data, new_vals)
            cache_db_engine = None
        return additional

    @classmethod
    def sql_filter(cls, sql):
        for s in AddressRepairBehavior.__sql_filter_arr:
            if sql.lower().find("delete") > -1 and sql.lower().find("isdeleted") > -1:
                continue
            if sql.lower().find(s) > -1:
                raise ValueError("检测到危险语句，停止执行")

    @classmethod
    def sql_filter_str(cls, sql):
        for s in AddressRepairBehavior.__sql_find_str_filter:
            sql = sql.replace(s, "")
        return sql

    @classmethod
    def get_engine(cls, key, db_config, cache_db_engine):
        """
        获得sql引擎
        """
        if (engine := cache_db_engine.get(key)) is None:
            pwd = db_config.get("pwd").replace("@", "%40")
            engine = create_engine('mssql+pymssql://%s:%s@%s/%s' % (db_config.get("user"),
                                                                    pwd,
                                                                    db_config.get("server"),
                                                                    db_config.get("database")),
                                   echo=False)
            engine = sessionmaker(bind=engine)
            cache_db_engine[key] = engine

        return engine

    @classmethod
    def get_db_data(cls, db_engine, data_vals, sql, column, value_column, mapping, try_match):

        cls.sql_filter(sql)
        handle_values = []



        for value in data_vals:
            value = cls.sql_filter_str(value)
            if try_match:
                many_match_result = []
                first_arr_val = value.split(',')[0]
                sql = sql.replace("${value}", first_arr_val)
            else:
                sql = sql.replace("${value}", value)


            with db_engine() as session:
                cursor = session.execute(sql)
                result = cursor.fetchall()

            is_all_match = False

            if len(result) > 1:
                for row in result:
                    # 全文匹配
                    for index in column:
                        if row[index] == value:
                            is_all_match = True
                            if mapping:
                                handle_values.append((value, str(row[value_column])))
                            else:
                                handle_values.append(str(row[value_column]))
                            break
                        if not is_all_match and try_match:
                            if row[index] == first_arr_val:
                                if mapping:
                                    many_match_result.append((value, str(row[value_column])))
                                else:
                                    many_match_result.append(str(row[value_column]))

                if try_match and len(many_match_result) == 1:
                    is_all_match = True
                    handle_values.append(many_match_result[0])
                    many_match_result.clear()

                if is_all_match:
                    break

            elif len(result) == 1:
                is_all_match = True
                if mapping is not None:
                    handle_values.append((value, str(result[0][value_column])))
                else:
                    handle_values.append(str(result[0][value_column]))
            if not is_all_match:
                handle_values.append(value)
        return handle_values


if __name__ == "__main__":
    test_data = {"test1": "NINGBO, CHINA", "test2": "DISHMAN-PHARMACEUTICAL-",
                  "address":
                      {"column": ["col1", "col2", "col3", "col4"],
                       "rows": [
                           ["aaedfkkk", "sc", "ITAAPS", "LIA NYU NGANG1, USA"],
                           ["12f", "HI", "mckdg", "IA"]
                       ]
                       }}
    AddressRepairBehavior().data_processing(
        None, test_data,
        [],
        {
            "address_repair": {
                "db": {
                    "pub": {"user": "co", "pwd": "Co&23@2332$22", "server": "db.dev.com:1433",
                            "database": "CO_PUB"}
                },
                "repairs": [
                    {"key": "test1", "table": "", "db_key": "pub",
                     "sql": "SELECT  [FullName],[LocalName],[name],[code],[Id] from Places WHERE IsDeleted = 0 and IsOcean = 1 and IsValid = 1 and ([FullName] like '%${value}%' or charindex([FullName],'${value}')> 0) ;",
                     "column": [0, 2, 3], "value": 4, "mapping": "youid",
                     "old_val_handle": "empty", "try_match": True}
                ]

            }
        }, None, {}
    )

    print(test_data)



# -*- coding: utf-8 -*-
from docparser.core.virtual_block import VirtualBlock


class TextVirtualBlock(VirtualBlock):
    """
    文本虚拟区块信息类
    """

    def __init__(self, sheet, block_config):
        VirtualBlock.__init__(self, sheet, block_config)



# -*- coding: utf-8 -*-
from marshmallow import fields
from docparser.config.excel.table_behavior_config import TableBehaviorConfigSchema
from docparser.config.excel.rule_config import RuleConfigSchema


class TableRuleConfigSchema(RuleConfigSchema):
    """
    键值对提取规则配置
    """
    # 表头
    column = fields.List(fields.Str(), required=True)
    # 行动组
    action = fields.List(fields.Nested(TableBehaviorConfigSchema, many=True))



from docparser.core.data_parser_base import DataParserBase
from docparser.core.tools import Tools
from docparser.core.behavior_data_parser import BehaviorDataParser
from docparser.config.excel import enums


class TableDataParser(DataParserBase):

    def check_config(self, config, key):
        """
        检查配置文件
        :param config: 配置
        :param key: 表名
        :return: 检查失败返回False，反之True
        """
        key = Tools.get_key(self.parent_key, key)
        err_count = len(self.errors)

        for k, v in config.items():
            if isinstance(v, str) and v == "":
                self.errors[Tools.get_key(key, k)] = r"<key: %s; 配置错误，不能为空>" % k

        if "column" not in config or len(config["column"]) == 0:
            self.errors[Tools.get_key(key, "column")] = r"<key: column; 表头至少需要一个列>"
        if "position_pattern" not in config:
            self.errors[Tools.get_key(key, "position_pattern")] = r"<key: position_pattern; 缺少配置>"
        if "behaviors" not in config:
            self.errors[Tools.get_key(key, "behaviors")] = r"<key: behaviors; 缺少配置>"
        for i, _config in enumerate(config["behaviors"]):
            if "over_action" not in _config:
                self.errors[Tools.get_key(key, "behaviors", "over_action")] = r"<key: over_action; 缺少配置>"
            if "value_pattern" not in _config:
                self.errors[Tools.get_key(key, "behaviors", "value_pattern")] = r"<key: value_pattern; 缺少配置>"
            if "value_format" in _config and Tools.check_regex_group(_config["value_format"]):
                self.errors[Tools.get_key(key, "behaviors", "value_format")] = r"<key: value_format; 正则表达式需要捕获组>"
            else:
                for j, _re in enumerate(_config["value_pattern"]):
                    if _config["over_action"] != enums.OverAction.row.name and Tools.check_regex_group(_re):
                        self.errors[Tools.get_key(key, "behaviors", "value_pattern", str(j),
                                                  _config["over_action"])] = r"<pattern: %s; 正则表达式需要捕获组>" % _re
                    elif _config["over_action"] == enums.OverAction.row.name and len(
                            config["column"]) > Tools.get_regex_group_count(_re):
                        self.errors[Tools.get_key(key, "behaviors", "value_pattern", str(j),
                                                  "row")] = r"<pattern: %s; 正则表达式的命名组格式不正确，或者数量与列数量不一致>" % _re

        return len(self.errors) == err_count

    def _parse(self, config, key):
        """
        解析表格数据
        :param config: 表格解析配置
        :param key: 表名
        :return:
        """
        regex = Tools.init_regex(config["position_pattern"])
        row_index, col_index = self._fixed_position_key(regex)
        if row_index == -1:
            self.errors["%s.%s" % (self.parent_key, key)] = r"<key:%s; 没有定位到有效位置>" % key
            return

        column = config["column"]
        separator = config["separator"]
        find_mode = config["find_mode"]
        separator_mode = config["separator_mode"]

        # 初始化表格
        table_key = key.split(".")[1]
        self.data[table_key] = {"column": column, "rows": []}
        behavior = BehaviorDataParser(config["behaviors"], row_index, col_index, separator, find_mode, separator_mode,
                                      table_key, "%s.%s" % (self.parent_key, key))
        behavior.match(self.data, self.errors, self.sheet)
        if len(self.data[table_key]["rows"]) > 0 and find_mode == enums.FindMode.v.name:
            self.data[table_key]["rows"].pop(0)



from docparser.core.tools import Tools


class DataParserBase:

    def __init__(self, sheet, data, errors, parent_key):
        """
        初始化表格解析器
        :param sheet: 被解析的表单
        """
        self.sheet = sheet
        self.data = data
        self.errors = errors
        self.parent_key = parent_key

    def check_config(self, config):
        pass

    def _callback_action_check(self, key, value, dic, parent_key, *args):
        if key == "pattern_list":
            for pattern in value:
                if Tools.check_regex_group(pattern):
                    self.errors[Tools.get_key(self.parent_key,
                                              parent_key)] = "<key: pattern_list; regex: %s; 正则没有捕获组>" % pattern
        if key in "pattern" and not isinstance(value, list):
            dic[key] = [value]

    def _check_config_action(self, config):
        Tools.recursion_dict(config, self._callback_action_check, None)

    def parse(self, config, key):
        self._prev(config, key)
        self._parse(config, key)
        self._after(config, key)
        return True

    def _parse(self, config, key):
        pass

    def _prev(self, config, key):
        pass

    def _after(self, config, key):
        pass

    def _fixed_position_key(self, pattern_list: list, start_row_index=0, start_col_index=0) -> (int, int):
        """
        定位键位置
        :param pattern_list: 定位正则表达式列表
        :return:
        """
        is_first = False
        for row_index in range(start_row_index, len(self.sheet)):
            for col_index in range(len(self.sheet[row_index])):
                # 防止重复匹配时，查询重复的位置
                if start_row_index == row_index and start_col_index == col_index:
                    is_first = True
                if is_first:
                    if Tools.match_find(self.sheet[row_index][col_index], pattern_list):
                        return row_index, col_index
        return -1, -1



import openpyxl
import os

from docparser.core.document_parser_base import DocumentParserBase


class ExcelDocumentParser(DocumentParserBase):
    """
    Excel文档解析器
    """
    def __init__(self, file, configs):
        """
        初始化
        :param file:文件路径
        :param configs: 配置
        """
        self._file = file
        self._configs = configs

        if not os.path.exists(file):
            raise FileNotFoundError

        work_book = openpyxl.load_workbook(file, read_only=True)
        self._sheet = work_book.worksheets[0]

    def parse(self):
        """
        根据配置抽取数据
        :return: 返回抽取的数据
        """

        data = {}
        errors = {}
        for key in self._configs.keys():
            item_config = self._configs[key]

            if item_config["type"] == 'table':
                errs = self._pre_check_and_process_table_config(item_config)
                if len(errs.keys()) > 0:
                    errors[key] = errs
                else:
                    data[key] = self._extract_table(item_config)
            else:
                text, errs = self._extract_text(item_config)
                if errs and len(errs.keys()) > 0:
                    errors[key] = errs
                else:
                    data[key] = text

        return data, errors

    def _extract_text(self, text_config) -> object:
        """
        提取文本域数据
        :param text_config:文本域配置
        :return: 返回文本域配置提取的内容
        """
        # text_block_extractor = ExcelTextBlockExtractor(self._sheet, text_config)
        # errs = text_block_extractor.check()
        # if len(errs.keys()) > 0:
        #     return "", errs
        #
        # text = text_block_extractor.extract()
        # text = self._text_process(text, text_config)
        #
        # return text, None

    def _extract_table(self, table_config) -> object:
        """
        提取表格数据
        :param table_config: 表格配置
        :return: 返回配置提取表格的数据
        """

        # rect = table_config["rect"]
        # items = []
        # original_columns_maps = table_config["original_columns_maps"]
        # extract_columns = table_config["extract_columns"]
        # top = original_columns_maps[list(original_columns_maps.keys())[0]]["row"] + 1
        #
        # row_spans = rect["bottom"] - rect["top"]
        # for row in range(top, rect["top"] + row_spans):
        #     item = {}
        #     for col_name in extract_columns.keys():
        #         col_title = extract_columns[col_name]["title"]
        #         col_map = original_columns_maps[col_title]
        #         text = ''
        #         for col in range(col_map["col"], col_map["col"] + col_map["span"]):
        #             text = text + ExcelExtractorUtils.get_sheet_cell_value(self._sheet, row,
        #                                                                    original_columns_maps[col_title]["col"])
        #
        #         # 设置单元格值
        #         item[col_name] = self._cell_value_process(text, row, col, table_config["extract_columns"][col_name])
        #
        #     # 添加表格行数据
        #     items.append(item)
        #
        # return items

    def _text_process(self, text, text_config):
        """
        文本结果处理
        :param text:区域文本数据
        :param text_config: 配置
        :return:
        """
        if text is None or len(text) == 0:
            return ""

        return text

    def _cell_value_process(self, text, row, col, col_config):
        """
        文本结果处理
        :param text:区域文本数据
        :param row:行
        :param col:列
        :param col_config: 列配置
        :return:
        """
        if text is None or len(text) == 0:
            return ""

        return text


if __name__ == '__main__':
    converter = ExcelDocumentParser(
        r"C:\Users\86134\Desktop\projects\email\653\cmacgm_noticeofarrival_csclyellowsea_0bh9ew1ma-at210727102222_769469_000019.xlsx",
        {
            # "vessel_name": {
            #     "type": "text",
            #     "rect": {
            #         "left_keyword": "VESSEL:",
            #         "right_keyword": "VOYAGE:",
            #         "bottom_keyword": "OPERATIONAL DISCH. PORT: PLACE"
            #     },
            #     "pattern": ".*"
            # },
            # "pod_eta": {
            #     "type": "text",
            #     "rect": {
            #         "left_keyword": "POD ETA:",
            #         "bottom_keyword": "FPD ETA:"
            #     },
            #     "pattern": ".*"
            # },
            # "it_number:": {
            #     "type": "text",
            #     "rect": {
            #         "left_keyword": "IT NUMBER:",
            #         "right_keyword": "PLACE OF ISSUE:",
            #     },
            #     "pattern": ".*"
            # },
            # "it_issued_date:": {
            #     "type": "text",
            #     "rect": {
            #         "left_keyword": "IT ISSUED DATE:",
            #     },
            #     "pattern": ".*"
            # },
            # "firms_code": {
            #     "type": "text",
            #     "rect": {
            #         "left_keyword": "FIRMS CODE:",
            #     },
            #     "pattern": ".*"
            # },
            # "shipper": {
            #     "type": "text",
            #     "rect": {
            #         "top_keyword": "SHIPPER",
            #         "bottom_keyword": "PLEASE NOTE :",
            #     },
            #     "pattern": ".*"
            # }

            "containers": {
                "type": "table",
                "extractor": "mixed",
                "max_rows": 1,
                "row_split_ref_col_name": "container_no",
                "col_split_chars": "  ",
                "rect": {
                    "top": {
                        "keyword": "CONTAINER  # ",
                        "include": True
                    },
                    "bottom": {
                        "keyword": "PLEASE NOTE :",
                    }
                },
                "columns": [
                    {
                        "name": "container_no",
                        "title": "CONTAINER #",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\w{0,20}",
                    }, {
                        "name": "seal_no",
                        "title": "SEAL #",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\w{0,20}",
                    }, {
                        "name": "container_size_type",
                        "title": "SIZE/TYPE #",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\d{1,10}\\s{1,2}\\[a-z|A-Z]{2,5}",
                    }, {
                        "name": "weight",
                        "title": "WEIGHT",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\d{0,10}",
                    }, {
                        "name": "measure",
                        "title": "MEASURE",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\w{0,5}",
                    }, {
                        "name": "free_business_last_free",
                        "title": "FREE BUSINESS LAST FREE",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "childrens": [
                            {
                                "name": "day_at_port",
                                "title": "DAYS AT PORT",
                                "title_h_align": "center",
                                "title_v_align": "middle",
                                "content_pattern": "\\w{0,20}",
                            },
                            {
                                "name": "day_at_ramp",
                                "title": "DAY AT RAMP",
                                "title_h_align": "center",
                                "title_v_align": "middle",
                                "content_pattern": "\\d{1,2}/\\d{1,2}/\\d{1,2}",
                            }
                        ]
                    }, {
                        "name": "pickup_no",
                        "title": "PICKUP #",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\w{0,20}",
                    },
                ]
            }
        })
    data = converter.extract()
    print(data)



import re

from docparser.core.tools import Tools
from docparser.config.excel import enums


class ActionDataParser:
    """
    行动解析器
    """

    def __init__(self, config, original_value, value, parent_key):
        self.config = config
        self.original_value = original_value
        self.value = value
        self.parent_key = parent_key
        self.keyword_list = config["keyword"]
        self.keyword = self.keyword_list[0] if isinstance(self.keyword_list, list) else self.keyword_list
        self.key = config["key"]
        self.action_type = "add" if "action_type" not in config else config["action_type"]
        self.pattern_list = [] if "pattern_list" not in config else config["pattern_list"]
        self.default = None if "default_value" not in config else config["default_value"]

    def __error(self, errors, reason=None):
        """
        组装错误信息
        :param errors:
        :param reason:
        :return:
        """
        error_message = "<keyword: %s; key: %s; type: %s; message:[%s]处理失败!%s;>" % (
            self.keyword, self.key, self.action_type, self.original_value, (reason or ""))
        key = "%s.%s" % (self.parent_key, self.keyword)
        errors[key] = error_message

    def __keyword_fixed_table(self, data):
        """
        通过keywor寻找表格单元格
        :param data: 结果字典
        :return: 行号，列号, 失败原因
        """
        arr = self.keyword.split("-")

        if len(arr) < 3:
            return -1, -1, "寻找表格位置失败,keyword的配置规则应为table-键-列号-行号,table-键-列号"

        table_key = arr[1]
        if table_key in data:
            column = data["column"]
            if len(column) >= int(arr[2]):
                other_row = data[table_key]["rows"]
                col_number = int(arr[2]) - 1
                if len(arr) == 4 and len(other_row) > 0:
                    row_number = int(arr[3]) - 1
                else:
                    row_number = -1
                return row_number, col_number, None
            else:
                return -1, -1, "被寻找的表格不符合配置的查找条件"

    def __loop_table(self, value, data, row_index, col_index, mode=True):
        """
        遍历表格
        :param value: 处理之后的原始值
        :param data: 结果字典中的具体表格
        :param row_index: 行索引
        :param col_index: 列索引
        :param mode: 添加模式，True：默认为覆盖模式， False: 追加模式
        :return:
        """
        if row_index > -1 and data[row_index][col_index].strip() == "":
            data[row_index][col_index] = value if mode else "%s%s" % (data[row_index][col_index], value)
        else:
            for i in range(len(data)):
                if len(data[i]) > col_index:
                    data[i][col_index] = value if mode else "%s%s" % (data[i][col_index], value)

    def __completion_table(self, data, value, values):
        """
        根据列值补全表格数据
        :param data: 结果字典
        :param value: 原始值
        :param values: 处理之后列值列表
        :return:
        """
        table_name = self.parent_key.split('.')[2]
        if table_name in data:
            col_index = int(self.key.split('_')[1]) - 1
            rows = data[table_name]["rows"]
            start_repl = False
            for i in range(1, len(rows)):
                if not start_repl and rows[i][col_index] in value:
                    start_repl = True
                if len(values) == 0:
                    return
                if start_repl:
                    rows[i][col_index] = values.pop(0)
        else:
            return "父级键的层级错误{%s}，没找到表格{%s}" % (self.parent_key, table_name)

    def change(self, data, errors, group_dict):
        """
        根据配置修改值
        :param data: 结果字典
        :param errors: 错误列表
        :param group_dict: 当前已完成匹配的捕获组
        :return:
        """
        value = ""
        if len(self.pattern_list) > 0:
            r_list = Tools.init_regex(self.pattern_list)
            gmatch = Tools.match_value(self.value, r_list)
            if gmatch:
                value = gmatch["value"]
        elif group_dict and self.key in group_dict:
            value = group_dict[self.key]
        else:
            value = self.value

        if self.default is not None and value is None:
            value = self.default
        elif value is None:
            self.__error(errors)

        row_index = col_index = -1
        if "table-" in self.keyword:
            row_index, col_index, msg = self.__keyword_fixed_table(data)
            if col_index == -1:
                self.__error(errors, msg)
                return
            else:
                data = data[self.keyword.split('-')[1]]["rows"]
        elif self.keyword in data and "column" in data[self.keyword]:
            self.__error(errors, "键值不能覆盖表格")
            return

        reason = None
        if self.action_type == enums.ActionType.add.name:
            self._add_mode(value, data, row_index, col_index)
        elif self.action_type == enums.ActionType.append.name:
            self._append_mode(value, data, row_index, col_index)
        elif self.action_type == enums.ActionType.cut.name:
            self._cut_mode(value, data)
        elif self.action_type == enums.ActionType.split.name:
            self._split_mode(value, data)
        if reason:
            self.__error(errors, reason)

    def _add_mode(self, value, data, row_index, col_index):
        """
        覆盖模式， 直接覆盖指定位置的值，没有则创建。 在表格模式下只覆盖空值列
        :param value: 处理之后的原始值
        :param data: 结果字典中的具体表格
        :param row_index: 行索引
        :param col_index: 列索引
        :return:
        """
        if col_index > -1:
            self.__loop_table(value, data, row_index, col_index, True)
        else:
            data.update({self.keyword: data[self.keyword] if value == "" and self.keyword in data else value})

    def _append_mode(self, value, data, row_index, col_index):
        """
        追加模式
        :param value: 处理之后的原始值
        :param data: 结果字典中的具体表格
        :param row_index: 行索引
        :param col_index: 列索引
        :return:
        """
        if col_index > -1:
            self.__loop_table(value, data, row_index, col_index, False)
        else:
            value = ("%s%s" % (data[self.keyword], value)) if self.keyword in data else value
            data.update({self.keyword: value})

    def _cut_mode(self, value, data):
        """
        裁剪模式
        :param value: 处理之后的原始值
        :param data: 结果字典
        :return:
        """
        if value.strip() != "":
            values = Tools.cut_value_to_string(value, self.keyword_list)
            self.__completion_table(data, self.value, values)

    def _split_mode(self, value, data):
        """
        分割模式
        :param value: 处理之后的原始值
        :param data: 结果字典
        :return:
        """
        if value.strip() != "":
            values = re.split(r"[%s]" % re.escape("".join(self.keyword_list)), value, re.I)
            self.__completion_table(data, self.value, values)



import re
from docparser.core.tools import Tools
from docparser.config.excel import enums
from docparser.core.action_data_parser import ActionDataParser


class BehaviorDataParser:
    """
    表格行为解析器
    """

    def __init__(self, config, row_start_index, col_start_index, separator, find_mode, separator_mode, table_key,
                 parent_key):
        """
        初始化
        :param config: 行为配置
        :param row_start_index: 匹配开始行索引
        :param col_start_index: 匹配开始列索引
        :param separator: 分割符
        :param find_mode: 查找方向
        :param separator_mode: 取值方式
        :param table_key: 表名
        :param parent_key 父级键
        """
        self.configs = config
        for i in range(len(self.configs)):
            v = self.configs[i]
            v["re"] = Tools.init_regex(v["value_pattern"])
            if "value_format" in v:
                v["re_format"] = Tools.init_regex(v["value_format"])
        self.repl_separator = " " if "repl_separator" not in config else config["repl_separator"]
        self.row_start = row_start_index
        self.col_start = col_start_index
        self.separator = separator
        self.find_mode = find_mode
        self.separator_mode = separator_mode
        self.table_key = table_key
        self.parent_key = parent_key

    def match(self, data, errors, sheet):
        """
        执行赋值行为
        :param data: 结果字典
        :param errors: 错误字典
        :param sheet: 表单
        :return:
        """
        row_start = self.row_start
        col_start = self.col_start
        table = data[self.table_key]
        index = 0
        config = self.configs[index]
        is_break = False
        for i in range(row_start, len(sheet)):
            for j in range(col_start, len(sheet[i])):
                if is_break and j > 0:
                    is_break = False
                    break
                for x in range(len(self.configs)):
                    if self.find_mode != enums.FindMode.v.name:
                        state = self.__horizontal(config, table, data, errors, sheet[i])
                    else:
                        state = self.__vertical(config, table, data, errors, sheet[i][j], j)

                    if state == enums.OverAction.end.name:
                        return
                    elif self.find_mode == enums.FindMode.v.name and state is not None:
                        break
                    elif state == enums.OverAction.skip.name or self.find_mode != enums.FindMode.v.name:
                        is_break = True
                        break
                    elif state is None:
                        index = 0 if index + 1 >= len(self.configs) else index + 1
                        config = self.configs[index]

    def __horizontal(self, config, table, data, errors, row):
        """
        横列模式解析
        :param config: 解析配置
        :param table: 结果字典中的目标表格
        :param data: 结果字典
        :param errors: 错误列表
        :param row: 被搜索的行
        :return:
        """
        value = self.separator.join(row)
        group_dict = None
        if "loop" in config and config["loop"] == 1 and config["over_action"] == enums.OverAction.row.name:
            group_dict = Tools.match_find_all(value, config["re"])
            if group_dict:
                for i in range(len(group_dict)):
                    if not isinstance(group_dict[i], str):
                        table["rows"].append(list(group_dict[i]))
                    else:
                        table["rows"].append([group_dict[i]])
                    for action in config["action"]:
                        if isinstance(action["key"], int):
                            ActionDataParser(action, group_dict[action["key"]], value,
                                             "%s.Behavior.%s" % (self.parent_key, config["over_action"])).change(data,
                                                                                                                 errors,
                                                                                                                 group_dict)
            else:
                return None
        else:
            group_dict, groups = Tools.match_value2(value, config["re"])
            if group_dict:
                if config["over_action"] == enums.OverAction.row.name:
                    row = []
                    for i in range(len(table["column"])):
                        row.append(group_dict["col_%s" % (i + 1)])
                    table["rows"].append(row)
                    for action in config["action"]:
                        ActionDataParser(action, group_dict[action["key"]], value,
                                         "%s.Behavior.%s" % (self.parent_key, config["over_action"])).change(data,
                                                                                                             errors,
                                                                                                             group_dict)
            else:
                return None

        return config["over_action"]

    def __vertical(self, config, table, data, errors, col, find_col_index):
        """
        纵列模式
        :param config: 解析配置
        :param table: 结果字典中的目标表格
        :param data: 结果字典
        :param errors: 错误列表
        :param col: 被搜索的列
        :param find_col_index: 若配置正则没有搜索到正确列索引，即列值为空时，则认为是被搜索的列索引是当前列索引
        :return:
        """

        group_dict, groups = Tools.match_value2(col, config["re"])
        if group_dict:
            if config["over_action"] == enums.OverAction.row.name:
                col_index = col_insert_index = 0
                col_list = []
                for k, v in group_dict.items():
                    if v and v.strip() != "":
                        col_index = int(k.split("_")[1]) - 1
                        v = self.__format_col(v, config)
                        col_list = v.split(self.separator)
                        break

                if len(col_list) == 0:
                    return None

                # 如果找到空匹配，默认当前列索引
                # if len(col_list) == 0:
                #     col_list.append("")
                #     col_index = find_col_index

                rows = table["rows"]
                row_count = len(rows)
                col_count = len(table["column"])

                # 首次添加
                if row_count == 0:
                    rows.append([col_index])
                    for i in range(len(col_list)):
                        rows.append([col_list[i]])
                else:
                    # 找哪一行缺列
                    # 缺列行的索引
                    row_insert_index = 0
                    # 有几列
                    incomplete_len = 0
                    for i in range(row_count):
                        if col_count > len(rows[i]):
                            incomplete_len = len(rows[i])
                            row_insert_index = i
                            break
                    # 首行加标记列
                    if row_insert_index == 0:
                        col_list.insert(0, col_index)

                    # 当前列数量
                    now_col_count = len(col_list)

                    # 找到当前列所在位置
                    for i in range(len(rows[0])):
                        if col_index > rows[0][i]:
                            col_insert_index = i + 1
                        elif col_index == rows[0][i]:
                            col_insert_index = i

                    # 当前准备赋值的列与拥有最大行数的列比较，得到需要补齐的行数.
                    diff = row_count - row_insert_index - now_col_count
                    # 存在缺列行才需要补

                    if incomplete_len > 0:
                        for i in range(abs(diff)):
                            if diff > 0:
                                # 当前列小于其他列长度，补全自己
                                col_list.append("")
                            else:
                                # 当前列大于其他列长度，补全行
                                rows.append([])

                    for i in range(row_insert_index, len(rows)):
                        # 对齐缺行列
                        for j in range(incomplete_len):
                            if len(rows[i]) < incomplete_len:
                                rows[i].append("")
                        rows[i].insert(col_insert_index, col_list.pop(0))

                for action in config["action"]:
                    ActionDataParser(action, group_dict[action["key"]], col,
                                     "%s.Behavior.%s" % (self.parent_key, config["over_action"])).change(data, errors,
                                                                                                         group_dict)
            return config["over_action"]
        elif groups:
            return config["over_action"]
        return None

    def __format_col(self, value, config):
        """
        格式化
        :param value: 捕获之后的纵列值
        :param config: 配置
        :return: 格式化后的字符串
        """
        if "re_format" in config:
            arr = config["re_format"].split(value)
            arr = [s.replace(self.separator, self.repl_separator) for s in arr if
                   s != "" and s != self.separator and s != self.repl_separator]
            value = self.separator.join(arr)
        return value



import re
import types


class Tools:
    """
    解析工具类
    """

    regex_check = re.compile(r'.*?[^(]*?\((?!\?[!#<>=(:])[^)]*?\)', re.I)
    regex_check_group = re.compile(r'.*?[^(]*?\(\?P<col_\d+>(?!\?[!#<=(:])[^)]*?\)', re.I)

    @staticmethod
    def init_regex(pattern_list):
        """
        预编译正则
        :param pattern_list: 正则列表
        :return:
        """
        if isinstance(pattern_list, list):
            if pattern_list and len(pattern_list) > 0:
                return [re.compile(regex, re.I) for regex in pattern_list]
            return []
        else:
            return re.compile(pattern_list, re.I)

    @staticmethod
    def match_value(value, pattern_list, default_value=None):
        """
        根据指定的正则表示列表,顺序解析指定的字符串,匹配成功后停止继续往下匹配.
        :param value: 需要匹配的字符串
        :param pattern_list: 正则表达式列表
        :param default_value: 解析失败后,返回默认值. 如果没有给默认参数,则输出None
        :return: 给定正则表达存在命名捕获组,返回dict.非命名捕获组,返回list
        """

        if len(value) > 0:
            for i, pattern in enumerate(pattern_list):
                match = pattern.match(value)
                if match:
                    group_list = match.groups()
                    group_dict = match.groupdict()
                    return group_dict if len(group_dict) > 0 else group_list
                else:
                    continue

        return default_value

    @staticmethod
    def match_value2(value, pattern_list):
        """
        根据指定的正则表示列表,顺序解析指定的字符串,匹配成功后停止继续往下匹配.
        :param value: 需要匹配的字符串
        :param pattern_list: 正则表达式列表
        :return: 同时返回dict，list
        """

        if len(value) > 0:
            for i, pattern in enumerate(pattern_list):
                match = pattern.match(value)
                if match:
                    group_list = match.groups()
                    group_dict = match.groupdict()
                    return group_dict, group_list
                else:
                    continue

        return None, None

    @staticmethod
    def match_find_all(value, pattern_list):
        """
        根据指定的正则表示列表,顺序解析指定的字符串,匹配成功后停止继续往下匹配.
        :param value: 需要匹配的字符串
        :param pattern_list: 正则表达式列表
        :return: 同时返回list(tuple)
        """

        if len(value) > 0:
            for i, pattern in enumerate(pattern_list):
                group_dict = pattern.findall(value)
                if group_dict is not None and len(group_dict) > 0:
                    return group_dict
                else:
                    continue
        return None

    @staticmethod
    def match_find(value, pattern_list):
        """
        根据指定正则表示列表,顺序解析指定的字符串,匹配成功后停止继续往下匹配.
        :param value: 需要匹配的字符串
        :param pattern_list: 正则表达式列表
        :return: 任意正则表达式匹成功,返回被捕获字符串.反之,返回None.
        """
        if len(value) > 0:
            for i, pattern in enumerate(pattern_list):
                match = pattern.match(value)
                if match:
                    return match.group(0)
        return False

    @staticmethod
    def table_data_next(data, row_index, col_index, next_cell=True):
        """
        获得表格下一个数据
        :param data: 源表格
        :param row_index: 查找开始的行
        :param col_index: 查找开始的列
        :param next_cell: True：下一个单元格，如果遇到行尾，则进去下一行的首列索引； False：仅找一下行的索引,列不变，如果下一行列索引超出，返回末尾列索引
        :return: 行索引，列索引 如果没有下一格则返回-1，-1
        """
        max_row_index = len(data) - 1

        if max_row_index >= row_index + 1 and not next_cell:
            max_col_index = len(data[row_index + 1]) - 1
            if max_col_index >= col_index:
                return row_index + 1, col_index
            else:
                return row_index + 1, len(data[row_index + 1]) - 1

        max_col_index = len(data[row_index]) - 1
        if max_col_index >= col_index + 1:
            return row_index, col_index + 1
        elif max_row_index >= row_index + 1:
            return row_index + 1, 0

        return -1, -1

    @staticmethod
    def cut_value_to_string(value, cut_list):
        """
        裁剪字符串
        :param value: 被裁剪的字符串
        :param cut_list: 固定值列表
        :return: 返回列表
        """

        arr = []
        value = value.lstrip()
        j = len(cut_list)
        index = 0

        for i in range(len(value)):
            _index = value.find(cut_list[index])
            if _index == 0:
                arr.append(cut_list[index])
                value = value[len(cut_list[index]):].rstrip().lstrip()
            else:
                index = 0 if index + 1 >= j else index + 1
            if len(value) == 0:
                break
        return arr

    @staticmethod
    def merge_cell(col_list):
        """
        合并列数据
        :param col_list: 列数据
        :return: 合并之后的二维列表
        """
        result = []
        max_row_size = max(len(i) for i in col_list)
        max_col_size = len(col_list)

        for j in range(max_row_size):
            lst = []
            for i in range(max_col_size):
                lst.append(col_list[i][j] if j < len(col_list[i]) else '')
            result.append(lst)

        return result

    @staticmethod
    def recursion_dict(dic, func, _type: type, parent_key="", *parameters):
        """
        递归字典，并对特定类型的值进行回调
        :param dic: 数据字典
        :param func: 回调函数 包含参数：当前键，当前值，父级对象，层级key,不定长参数
        :param _type: 特定类型
        :param parent_key: 父级的键 默认空字符串
        :param parameters: 不定长参数，默认传回回调函数
        :return: 无
        """
        if not isinstance(dic, dict) or (
                not isinstance(func, types.FunctionType) and not isinstance(func, types.MethodType)):
            return

        keys_list = list(dic.keys())

        for index in range(len(keys_list)):
            k = keys_list[index]
            v = dic[keys_list[index]]

            if _type is None or isinstance(v, _type):
                func(k, v, dic, parent_key, *parameters)
            if _type is None or not isinstance(v, _type):
                if isinstance(v, list):
                    for i in range(len(v)):
                        if isinstance(v[i], dict):
                            Tools.recursion_dict(v[i], func, _type, "%s.%s.%d" % (parent_key, k, i), *parameters)
                elif isinstance(v, dict):
                    Tools.recursion_dict(v, func, _type, "%s.%s" % (parent_key, k), *parameters)

    @staticmethod
    def check_regex_group(regex_str):
        """
        检查正则表达式是否存在捕获组
        :param regex_str: 正则表达式字符串 或 正则表达式列表
        :return: 存在返回False,反之返回True
        """
        if isinstance(regex_str, list):
            return not all(Tools.regex_check.match(_str) for _str in regex_str)
        return Tools.regex_check.match(regex_str) is None

    @staticmethod
    def get_regex_group_count(regex_str):
        """
        返回正则表达式中带有【col_数字】命名捕获组的数量
        :param regex_str: 正则表达式字符串
        :return: 存在返回True,反之返回False
        """

        match_all = Tools.regex_check_group.findall(regex_str)
        return len(match_all)

    @staticmethod
    def len_table_cell_count(table):
        """
        获得二维列表的单元格的数量
        :param table: 二维列表
        :return:
        """
        cell_count = 0
        for i in range(len(table)):
            cell_count += len(table[i])
        return cell_count

    @staticmethod
    def get_key(*args):
        """
        组合键名
        :param args: 键组
        :return: 键
        """
        return ".".join(args)

    @staticmethod
    def get_dict_for_index(dic, index):
        if isinstance(dic, dict):
            for i, value in enumerate(dic.values()):
                if i == index:
                    return value
        else:
            return dic[index]



import re

from docparser.config.excel import enums
from docparser.core.tools import Tools
from docparser.core.data_parser_base import DataParserBase
from docparser.core.action_data_parser import ActionDataParser


class KvDataParser(DataParserBase):

    def check_config(self, config, key):
        """
        检查键值对配置文件
        :param config: 配置
        :param key: 键
        :return: 检查失败返回False，反之True
        """
        key = Tools.get_key(self.parent_key, key)
        self._check_config_action(config)

        err_count = len(self.errors)

        cell_count = Tools.len_table_cell_count(self.sheet)
        if "position_pattern" not in config:
            self.errors[Tools.get_key(key, "value_pattern")] = r"<key: value_pattern; 缺少配置或者为空>"
        else:
            for i, pattern in enumerate(config["position_pattern"]):
                if pattern.strip() == "":
                    self.errors[Tools.get_key(key, "value_pattern", str(i))] = r"<value_pattern不能为空>"
        if "repeat_count" in config:
            if config["repeat_count"] <= 0:
                config["repeat_count"] = 1
            elif config["repeat_count"] > cell_count:
                config["repeat_count"] = cell_count
        else:
            config["repeat_count"] = 1
        if "is_split_cell" in config and config["is_split_cell"] == 1:
            if Tools.check_regex_group(config["split_pattern"]):
                self.errors[Tools.get_key(key, "split_pattern")] = r"<key: split_pattern; 缺少分列的捕获组>"
        if "value_pattern" not in config:
            self.errors[Tools.get_key(key, "value_pattern")] = r"<key: value_pattern; 缺少配置>"
        else:
            for j, _re in enumerate(config["value_pattern"]):
                if config["separator_mode"] != enums.SeparatorMode.split.name and Tools.check_regex_group(_re):
                    self.errors[
                        Tools.get_key(key, "value_pattern", str(j))] = r"<key: value_pattern; regex: %s; 缺少捕获组>" % _re

        for i,action in enumerate(config["action"]):
            if "key" not in action:
                action["key"] = i

        return len(self.errors) == err_count

    def _parse(self, config, key):
        """
        解析键值对数据
        :param config: 键值对解析配置
        :param key: 键
        :return:
        """
        regex = Tools.init_regex(config["position_pattern"])
        row_index = col_index = 0
        for repl in range(config["repeat_count"]):
            row_index, col_index = self._fixed_position_key(regex, row_index, col_index)
            if row_index == -1:
                self.errors[Tools.get_key(self.parent_key, key, "fixed_position_key")] = r"<没有定位到有效位置>"
                return

            find_mode = config["find_mode"]
            separator_mode = config["separator_mode"]

            is_split_cell = config["is_split_cell"]
            split_pattern = config["split_pattern"]
            if is_split_cell == 1:
                self._split_cell(row_index, col_index, key, split_pattern)

            group_dict, value, row_index, col_index = self._fixed_position_value(find_mode, separator_mode,
                                                                                 config["value_pattern"], row_index,
                                                                                 col_index)
            if config["repeat_count"] > 1 and find_mode == enums.FindMode.default.name and row_index > -1:
                row_index, col_index = Tools.table_data_next(self.sheet, row_index, col_index)

            for i, action in enumerate(config["action"]):
                v = None
                if group_dict:
                    if "key" not in action:
                        action["key"] = i
                    if "action_type" not in action:
                        action["action_type"] = "add"

                    if isinstance(action["key"], str) and action["key"] in group_dict:
                        v = group_dict[action["key"]]
                    elif isinstance(action["key"], int) and action["key"] < len(group_dict):
                        v = Tools.get_dict_for_index(group_dict, action["key"])

                    if "table-" not in action["keyword"] and action["keyword"] not in self.data:
                        self.data.update({action["keyword"]: ""})
                v = v if v else ""
                action_parser = ActionDataParser(action, value, v, Tools.get_key(self.parent_key, key))
                action_parser.change(self.data, self.errors, group_dict)

    def _split_cell(self, row_index, col_index, key, split_pattern):
        """
        拆分单元
        :param row_index: 单元格所在行号
        :param col_index: 单元格所在列号
        :return:
        """
        group_dict = Tools.match_value(self.sheet[row_index][col_index], Tools.init_regex(split_pattern))
        if group_dict:
            arr = group_dict.values()
            self.sheet[row_index].pop(col_index)
            for i, v in enumerate(arr):
                self.sheet[row_index].insert(row_index + i, v)
        else:
            self.errors[Tools.get_key(self.parent_key, key, "split_cell")] = r"<没有定位到有效位置>"

    def _fixed_position_value(self, find_mode, separator_mode, value_pattern, start_row_index, start_col_index):
        """
        定位值
        :param find_mode: 查找方向
        :param separator_mode: 分割膜式
        :param value_pattern: 正则表达式组
        :param start_row_index 开始行索引
        :param start_col_index 开始列索引
        :return: 匹配值， 原始值
        """
        row_index = start_row_index
        col_index = start_col_index
        cell_value = None
        cell_count = Tools.len_table_cell_count(self.sheet)
        for i in range(cell_count):
            if (find_mode == enums.FindMode.default.name and i == 0) or (
                    find_mode != enums.FindMode.default.name and i > 0):
                value = self.sheet[row_index][col_index]
                if separator_mode == enums.SeparatorMode.split.name:
                    cell_value = re.split(r'[%s]' % "".join(value_pattern), value, re.I)
                else:
                    cell_value = Tools.match_value(value, Tools.init_regex(value_pattern))
            if cell_value:
                return cell_value, value, row_index, col_index
            else:
                mode = find_mode == enums.FindMode.h.name
                if find_mode != enums.FindMode.default.name:
                    row_index, col_index = Tools.table_data_next(self.sheet, row_index, col_index, mode)
                    if row_index == -1:
                        row_index = 0
                        col_index = 0
                else:
                    break

        return None, None, -1, -1



# -*- coding: utf-8 -*-
from marshmallow import fields
from docparser.config.excel.action_config import ActionConfigSchema
from docparser.config.excel.rule_config import RuleConfigSchema


class KvRuleConfigSchema(RuleConfigSchema):
    """
    键值对提取规则配置
    """
    # 提取规则组
    value_pattern = fields.List(fields.Str(), required=True)
    # 重复匹配次数
    repeat_count = fields.Integer(default=1)
    # 是否拆分单元格
    is_split_cell = fields.Integer(default=0)
    # 拆分规则
    split_pattern = fields.List(fields.Str())
    # 行动组
    action = fields.List(fields.Nested(ActionConfigSchema, many=True))



# -*- coding: utf-8 -*-
from docparser.core.table_virtual_block import TableVirtualBlock


class CmaAnTableVirtualBlock(TableVirtualBlock):
    """
    CMA到港通知书表格扩展实现
    """

    def __init__(self, sheet, block_config):
        self.virtual_table = None

   



class DocumentParserBase:
    """
    文档解析器基类
    """

    def parse(self, file, configs):
        """
        文档解析
        :param file:文件
        :param configs:提取配置
        :return: 无
        """
        pass



# -*- coding: utf-8 -*-
from marshmallow import Schema, fields
from docparser.config.excel.kv_rule_config import KvRuleConfigSchema
from docparser.config.excel.table_rule_config import TableRuleConfigSchema


class ExcelDocumentConfigSchema(Schema):
    """
    Excel 文档配置
    """
    # ID
    id = fields.Str(required=True)

    # 名称
    name = fields.Str(required=True)

    # 多页模式
    multi_page = fields.Integer(required=False, default=0)

    # 页头正则
    page_pattern = fields.List(fields.Str(), required=False)

    # 键值采集数据项配置集合
    kv = fields.Dict(Keys=fields.Str(), Values=fields.Nested(KvRuleConfigSchema))

    # 表格采集数据项配置集合
    table = fields.Dict(Keys=fields.Str(), Values=fields.Nested(TableRuleConfigSchema))


if __name__ == '__main__':

    config = {
        "id": "group_A",
        "name": "多页模板",
        "kv": {
            "Arrival Vessel": {
                "position_pattern": [r"^Arrival Vessel"],
                "value_pattern": [
                    r"Arrival Vessel\s*:\s*(\w*\s*\w*\s*\d+\s+\w+)\s{4,}[\w\W]*?\nB/L No\s*:\s*([\w]*?)\s{1,}"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Arrival Vessel"},
                    {"keyword": "B/L No"}
                ]
            },
            "ETA": {
                "position_pattern": [r"^ETA/ETB"],
                "value_pattern": [
                    r"[\w\W]*?(\d{2,}\s*[a-zA-Z]{3,}\s*\d{2,}\s*\d{2,}\:\d{2,}\([a-zA-Z]{2}\))[\w\W]*?(\d{2,}\s*[a-zA-Z]{3,}\s*\d{2,}\s*\d{2,}\:\d{2,})[\w\W]*?([a-zA-Z]{3,}\s*\d+\s*[a-zA-Z]{3,})"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "ETA"},
                    {"keyword": "vailable Date"},
                    {"keyword": "Port Free Time"},
                ]
            }
        },
        "table": {
            "bill": {
                "position_pattern": [r"^CONTAINER#"],
                "separator": "\n",
                "find_mode": "h",
                "separator_mode": "regex",
                "column": ["CHG", "RATED AS", "RATE", "PE", "COLLECT"],
                "behaviors": [
                    {
                        "over_action": "row",
                        "loop": 1,
                        "value_pattern": [
                            r"(?P<col_1>[a-zA-Z]*\s{1,}[a-zA-Z]*)\s{1,}(?P<col_2>\d{1,}\.\d{1,})\s{1,}(?P<col_3>\d{1,}\.\d{1,})\s*(?P<col_4>\w{1,})\s{1,}(?P<col_5>\d{1,}\.\d{1,})(?:\n|$)"],
                        "action": []
                    },
                    {
                        "over_action": "end",
                        "value_pattern": ["^(DESTINATION)"]
                    }
                ]
            }
        }
    }

    schema = ExcelDocumentConfigSchema()
    result = schema.load(config)
    print(result)



# -*- coding: utf-8 -*-
from marshmallow import Schema, fields
from docparser.config.excel.custom_fields import CustomFields
from docparser.config.excel.action_config import ActionConfigSchema


class TableBehaviorConfigSchema(Schema):
    """
    匹配行为组
    """
    # 匹配行为结束后之后的操作
    over_action = CustomFields.OverActionField(required=True)

    # 行匹配的正则
    row_pattern = fields.List(fields.Str(), required=True)

    # 格式化匹配后的值， 纵列模式在分列前格式化字符串，其余条件下则是在赋值之前格式化
    value_format = fields.Str(required=False)

    # 分隔符置换符
    repl_separator = fields.Str(required=False, default=0)

    # 循环模式
    loop = fields.Integer(required=False)

    # 行动组
    action = fields.List(fields.Nested(ActionConfigSchema, many=True))



import os
from lxml import etree

from docparser.core.document_parser_base import DocumentParserBase


class HtmlDocumentParser(DocumentParserBase):
    """
    Html文档解析器
    """

    def __init__(self, file, configs):
        """
        初始化
        :param file:文件路径
        :param configs: 配置
        """
        self._file = file
        self._configs = configs

        if not os.path.exists(file):
            raise FileNotFoundError

        self._html = etree.parse(file, etree.HTMLParser())

    def parse(self):
        """
        根据配置解析数据
        :return: 返回抽取的数据
        """

        data = {}
        errors = {}
        for key in self._configs.keys():
            item_config = self._configs[key]

            if item_config["type"] == 'table':
                errs = self._pre_check_and_process_table_config(item_config)
                if len(errs.keys()) > 0:
                    errors[key] = errs
                else:
                    data[key] = self._extract_table(item_config)
            else:
                text, errs = self._extract_text(item_config)
                if errs and len(errs.keys()) > 0:
                    errors[key] = errs
                else:
                    data[key] = text

        return data, errors


if __name__ == '__main__':
    converter = HtmlDocumentParser(
        r"C:\Users\86134\Desktop\projects\email\653\cmacgm_noticeofarrival_csclyellowsea_0bh9ew1ma-at210727102222_769469_000019.xlsx",
        {
            # "vessel_name": {
            #     "type": "text",
            #     "rect": {
            #         "left_keyword": "VESSEL:",
            #         "right_keyword": "VOYAGE:",
            #         "bottom_keyword": "OPERATIONAL DISCH. PORT: PLACE"
            #     },
            #     "pattern": ".*"
            # },
            # "pod_eta": {
            #     "type": "text",
            #     "rect": {
            #         "left_keyword": "POD ETA:",
            #         "bottom_keyword": "FPD ETA:"
            #     },
            #     "pattern": ".*"
            # },
            # "it_number:": {
            #     "type": "text",
            #     "rect": {
            #         "left_keyword": "IT NUMBER:",
            #         "right_keyword": "PLACE OF ISSUE:",
            #     },
            #     "pattern": ".*"
            # },
            # "it_issued_date:": {
            #     "type": "text",
            #     "rect": {
            #         "left_keyword": "IT ISSUED DATE:",
            #     },
            #     "pattern": ".*"
            # },
            # "firms_code": {
            #     "type": "text",
            #     "rect": {
            #         "left_keyword": "FIRMS CODE:",
            #     },
            #     "pattern": ".*"
            # },
            # "shipper": {
            #     "type": "text",
            #     "rect": {
            #         "top_keyword": "SHIPPER",
            #         "bottom_keyword": "PLEASE NOTE :",
            #     },
            #     "pattern": ".*"
            # }

            "containers": {
                "type": "table",
                "extractor": "mixed",
                "max_rows": 1,
                "row_split_ref_col_name": "container_no",
                "col_split_chars": "  ",
                "rect": {
                    "top": {
                        "keyword": "CONTAINER  # ",
                        "include": True
                    },
                    "bottom": {
                        "keyword": "PLEASE NOTE :",
                    }
                },
                "columns": [
                    {
                        "name": "container_no",
                        "title": "CONTAINER #",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\w{0,20}",
                    }, {
                        "name": "seal_no",
                        "title": "SEAL #",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\w{0,20}",
                    }, {
                        "name": "container_size_type",
                        "title": "SIZE/TYPE #",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\d{1,10}\\s{1,2}\\[a-z|A-Z]{2,5}",
                    }, {
                        "name": "weight",
                        "title": "WEIGHT",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\d{0,10}",
                    }, {
                        "name": "measure",
                        "title": "MEASURE",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\w{0,5}",
                    }, {
                        "name": "free_business_last_free",
                        "title": "FREE BUSINESS LAST FREE",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "childrens": [
                            {
                                "name": "day_at_port",
                                "title": "DAYS AT PORT",
                                "title_h_align": "center",
                                "title_v_align": "middle",
                                "content_pattern": "\\w{0,20}",
                            },
                            {
                                "name": "day_at_ramp",
                                "title": "DAY AT RAMP",
                                "title_h_align": "center",
                                "title_v_align": "middle",
                                "content_pattern": "\\d{1,2}/\\d{1,2}/\\d{1,2}",
                            }
                        ]
                    }, {
                        "name": "pickup_no",
                        "title": "PICKUP #",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\w{0,20}",
                    },
                ]
            }
        })
    data = converter.extract()
    print(data)



# -*- coding: utf-8 -*-
import importlib


class DocParserFactory:
    """
    文档解析器创建工厂
    """

    @staticmethod
    def create(doc_type, file, config):
        """
        根据类型创建对应类型的文档解析器

        """
        try:

            module_name = 'docparser.implements.%s_document_parser' % doc_type

            class_name = '%sDocumentParser' % doc_type.capitalize()
            virtual_block_module = importlib.import_module(module_name)
            cls = getattr(virtual_block_module, class_name)

            return cls(file, config)
        except ModuleNotFoundError:
            raise




# -*- coding: utf-8 -*-
from docparser.config.excel.enums import VerticalAlignMode
from docparser.core.utils import Utils



class Rect:
    """
    区域信息
    """

    def __init__(self, sheet, rect_config):
        # 表单对象
        self.sheet = sheet
        # 区域配置
        self.rect_config = rect_config
        # 区域定位信息
        self.positions = dict()
        # 区域内数据行集合(存放表单原始Cell值)
        self.rows = []
        # 开始行-Sheet内
        self.start_row_in_sheet = 0
        # 开始行-Cell内
        self.start_row_in_cell = 0
        # 结束行-Cell内
        self.end_row_in_sheet = 0
        # 结束行-Cell内
        self.end_row_in_cell = 0
        # 开始列-Sheet内
        self.start_col_in_sheet = 0
        # 开始列-Cell内
        self.start_col_in_cell = 0
        # 结束列-Sheet内
        self.end_col_in_sheet = 0
        # 结束列-Cell内
        self.end_col_in_cell = 0

        self.__init_positions()
        self.__init_rect()
        self.__init_rows()

    def __init_positions(self):
        """
        初始化定位信息
        """
        matcheds = {}

        match_times = 0
        for r in range(1, self.sheet.max_row):
            for c in range(1, self.sheet.max_column):
                # 获取单元格值
                cell_value, max_row_in_cell, max_col_in_cell = str(Utils.get_sheet_cell_info(self.sheet, r, c))
                if cell_value is None or len(cell_value.strip()) == 0:
                    continue

                for p in ["left", "right", "top", "bottom"]:
                    # 获取定位的关键字
                    position_config = self.rect_config[p] if p in self.rect_config else None
                    if position_config is None:
                        continue

                    # 未配置关键字则不匹配
                    keyword = position_config["keyword"]
                    if len(keyword) == 0:
                        continue

                    # 根据关键字匹配在单元格内部位置
                    start_row_in_cell, start_col_in_cell = Utils.location_in_cell_pos(
                        cell_value, keyword)
                    if start_row_in_cell > -1:
                        # 确定第几次匹配
                        if match_times < position_config['match_times'] - 1:
                            match_times = match_times + 1
                            continue

                        if p == "left":
                            self.__init_left_position(position_config, r, c, keyword)
                        elif p == 'right':
                            self.__init_right_position(position_config, r, c, keyword)
                        elif p == 'top':
                            self.__init_top_position(position_config, r, c, keyword)
                        elif p == 'bottom':
                            self.__init_bottom_position(position_config, r, c, keyword)

                        matcheds[p + "_match"] = True

        # 验证区域存在关键是否匹配请阿卡
        errors = self.__valid_positions(matcheds)
        if len(errors.keys()) == 0:
            self.__adjust_missing_positions()

        return errors

    def __init_rect(self):
        """
        初始化区域信息
        """
        self.start_row_in_sheet = 0
        self.end_row_in_sheet = self.sheet.max_row
        self.start_col_in_sheet = 1
        self.end_col_in_sheet = self.sheet.max_column

        if "top" in self.positions:
            top_pos = self.positions["top"]
            self.start_row_in_sheet = top_pos["row_in_sheet"]
            self.start_row_in_cell = top_pos["row_in_cell"] + 1

        if "bottom" in self.positions:
            bottom_pos = self.positions["bottom"]
            self.end_row_in_sheet = bottom_pos["row_in_sheet"]
            self.end_row_in_cell = bottom_pos["row_in_cell"] - 1

        if "left" in self.positions:
            left_pos = self.positions["left"]
            self.start_col_in_sheet = left_pos["col_in_sheet"]
            self.start_col_in_cell = left_pos["col_in_cell"]

        if "right" in self.positions:
            right_pos = self.positions["right"]
            self.start_col_in_sheet = right_pos["col_in_sheet"]
            self.start_col_in_cell = right_pos["col_in_cell"]

    def __init_rows(self):
        """
        初始化数据行信息
        """
        for r in range(self.start_row_in_sheet, self.end_row_in_sheet):
            row = []
            for c in range(self.start_col_in_sheet, self.end_col_in_sheet):
                row.append(self.sheet.cell(r, c).value)
            self.rows.append(row)

    def __init_left_position(self, position_config, row, col, keyword):
        """
        初始化左侧定位信息
        """

        cell_value, row_count_in_cell, col_count_in_cell = Utils.get_sheet_cell_info(self.sheet, row, col)
        start_row_in_cell, start_col_in_cell = Utils.location_in_cell_pos(cell_value, keyword)

        temp_row_in_sheet = row
        temp_col_in_sheet = col
        temp_row_in_cell = start_row_in_cell
        temp_col_in_cell = start_col_in_cell

        # 如果关键不在区域内,则开始位置需要重新计算
        if position_config['is_outer']:
            temp_col_in_cell = temp_col_in_cell + len(keyword)

            # 如果关键字已经在末尾，则区域计算开始跳到下一个单元格
            if temp_col_in_cell >= col_count_in_cell:
                temp_col_in_sheet = temp_col_in_sheet + 1
                temp_col_in_cell = 0

        position_info = {
            "row_in_sheet": temp_row_in_sheet,
            "col_in_sheet": temp_col_in_sheet,
            "row_in_cell": temp_row_in_cell,
            "col_in_cell": temp_col_in_cell,
            "row_count_in_cell": row_count_in_cell,
            "col_count_in_cell": col_count_in_cell
        }

        self.positions['left'] = position_info

        # 根据左侧定位及对齐推测出TOP定位信息
        if 'top' not in self.positions and 'v_align' in position_config and position_config['v_align'] == VerticalAlignMode.top.name:
            self.positions['top'] = position_info

        # 根据左侧定位及对齐推测出Bottom定位信息
        if 'bottom' not in self.positions and 'v_align' in position_config and position_config['v_align'] == VerticalAlignMode.bottom.name:
            self.positions['bottom'] = position_info

        return position_info

    def __init_right_position(self, position_config, row, col, keyword):
        """
        初始化右侧定位信息
        """

        # 根据关键字匹配在单元格内部位置
        cell_value, row_count_in_cell, col_count_in_cell = Utils.get_sheet_cell_info(self.sheet, row, col)
        start_row_in_cell, start_col_in_cell = Utils.location_in_cell_pos(cell_value, keyword)

        temp_row_in_sheet = row
        temp_col_in_sheet = col
        temp_row_in_cell = start_row_in_cell
        temp_col_in_cell = start_col_in_cell

        # 如果关键不在区域内,则开始位置需要重新计算
        if not position_config['is_outer']:
            temp_col_in_cell = temp_col_in_cell + len(keyword)
        else:
            # 如果关键字已经在开始，则区域计算开始跳到上一个单元格
            if temp_col_in_cell == 0:
                temp_col_in_sheet = temp_col_in_sheet - 1
                _, _, max_col_in_cell = Utils.get_sheet_cell_info(self.sheet, row, col - 1)
                temp_col_in_cell = max_col_in_cell

        position_info = {
            "row_in_sheet": temp_row_in_sheet,
            "col_in_sheet": temp_col_in_sheet,
            "row_in_cell": temp_row_in_cell,
            "col_in_cell": temp_col_in_cell,
            "row_count_in_cell": row_count_in_cell,
            "col_count_in_cell": col_count_in_cell
        }

        self.positions['right'] = position_info

        # 根据右侧定位及对齐推测出TOP定位信息
        if 'top' not in self.positions and 'v_align' in position_config and position_config['v_align'] == VerticalAlignMode.top.name:
            self.positions['top'] = position_info

        # 根据右侧定位及对齐推测出Bottom定位信息
        if 'bottom' not in self.positions and 'v_align' in position_config and position_config['v_align'] == VerticalAlignMode.bottom.name:
            self.positions['bottom'] = position_info

        return position_info

    def __init_top_position(self, position_config, row, col, keyword):
        """
        初始化顶侧定位信息
        """

        # 根据关键字匹配在单元格内部位置
        cell_value, row_count_in_cell, col_count_in_cell = Utils.get_sheet_cell_info(self.sheet, row, col)
        start_row_in_cell, start_col_in_cell = Utils.location_in_cell_pos(cell_value, keyword)

        temp_row_in_sheet = row
        temp_col_in_sheet = col
        temp_row_in_cell = start_row_in_cell
        temp_col_in_cell = start_col_in_cell

        # 如果关键不在区域内,则开始位置需要重新计算
        if position_config['is_outer']:
            temp_row_in_cell = temp_row_in_cell + 1

            # 如果关键字已经在末尾，则区域计算开始跳到下一个单元格
            if temp_row_in_cell == row_count_in_cell:
                temp_row_in_sheet = temp_row_in_sheet + 1
                temp_row_in_cell = 0

        position_info = {
            "row_in_sheet": temp_row_in_sheet,
            "col_in_sheet": temp_col_in_sheet,
            "row_in_cell": temp_row_in_cell,
            "col_in_cell": temp_col_in_cell,
            "row_count_in_cell": row_count_in_cell,
            "col_count_in_cell": col_count_in_cell
        }

        self.positions['top'] = position_info

        # 根据右侧定位及对齐推测出TOP定位信息
        if 'left' not in self.positions and 'h_align' in position_config and position_config['h_align'] == VerticalAlignMode.left.name:
            self.positions['left'] = position_info

        # 根据右侧定位及对齐推测出Bottom定位信息
        if 'right' not in self.positions and 'h_align' in position_config and position_config['h_align'] == VerticalAlignMode.right.name:
            self.positions['right'] = position_info

        return position_info

    def __init_bottom_position(self, position_config, row, col, keyword):
        """
        初始化低侧定位信息
        """

        # 根据关键字匹配在单元格内部位置
        cell_value, row_count_in_cell, col_count_in_cell = Utils.get_sheet_cell_info(self.sheet, row, col)
        start_row_in_cell, start_col_in_cell = Utils.location_in_cell_pos(cell_value, keyword)

        temp_row_in_sheet = row
        temp_col_in_sheet = col
        temp_row_in_cell = start_row_in_cell
        temp_col_in_cell = start_col_in_cell

        # 如果关键不在区域内,则开始位置需要重新计算
        if position_config['is_outer']:
            if temp_row_in_cell == 0:
                temp_row_in_sheet = temp_row_in_sheet - 1
                _, max_row_in_cell, _ = Utils.get_sheet_cell_info(self.sheet, row - 1, col)
                temp_row_in_cell = max_row_in_cell

        position_info = {
            "row_in_sheet": temp_row_in_sheet,
            "col_in_sheet": temp_col_in_sheet,
            "row_in_cell": temp_row_in_cell,
            "col_in_cell": temp_col_in_cell,
            "row_count_in_cell": row_count_in_cell,
            "col_count_in_cell": col_count_in_cell
        }
        self.positions['bottom'] = position_info

        # 根据右侧定位及对齐推测出TOP定位信息
        if 'left' not in self.positions and 'h_align' in position_config and position_config['h_align'] == VerticalAlignMode.left.name:
            self.positions['left'] = position_info

        # 根据右侧定位及对齐推测出Bottom定位信息
        if 'right' not in self.positions and 'h_align' in position_config and position_config['h_align'] == VerticalAlignMode.right.name:
            self.positions['right'] = position_info

        return position_info

    def __valid_positions(self, matcheds):
        """
        验证定位信息是否配置正确
        """
        errors = dict()
        for p in ["left", "right", "top", "bottom"]:
            keyword = self.rect_config[p]['keyword'] if p in self.rect_config else ""
            if len(keyword) > 0 and (p + "_match" not in matcheds):
                errors[p + "_keyword"] = "【%s】未在文档中出现,请检查配置！" % keyword

        return errors



# -*- coding: utf-8 -*-
from docparser.core.table_virtual_block import TableVirtualBlock


class MixedTableVirtualBlock(TableVirtualBlock):
    """
    混杂模式表格虚拟区块实现
    """

    def __init__(self, sheet, block_config):
        self.virtual_table = None

   



# -*- coding: utf-8 -*-
from docparser.core.rect import Rect


class VirtualBlock:
    """
    虚拟区块信息类
    """

    def __init__(self, sheet, block_config):
        # Excel表格
        self.sheet = sheet

        # 虚拟块配置信息
        self.block_config = block_config

        self.virtual_block_data = []

        self.rect = None
        # 初始化区块信息
        # self.__init_rect()

        # 初始化虚拟块数据
        # self.__init_virtual_block_data()

    def __init_rect(self):
        """
        初始化区域信息
        """
        self.rect = Rect(self.sheet, self.block_config)

    # def __init_virtual_block_data(self):
    #     """
    #     初始化虚拟块数据
    #     """
    #     self.virtual_block_data = []

    def _pre_build(self, rect):
        """
        数据前置处理
        """

        pass

    def _build(self):
        """
        输出结果
        """
        self.virtual_block_data = []
        return None

    def _post_build(self, data):
        """
        数据后置处理
        """
        return None

    def build(self):
        """
        输出结果
        """

        # 根据配置计算区域信息
        self.rect = Rect(self.sheet, self.block_config)

        # 预构建
        self._pre_build(self.rect)

        # 构建虚拟块规范数据
        self._build()

        # 构建后置处理
        self._post_build(self.virtual_block_data)

    def _pre_output(self, virtual_block_data):
        """
        输出前置处理
        """
        return virtual_block_data

    def _post_output(self, output_data):
        """
        输出后置处理
        """
        return output_data

    def _build_output(self, virtual_block_data):
        """
        构建输出结果
        """
        return virtual_block_data

    def output(self):
        """
        输出
        """

        # 前置处理虚拟块缓存数据
        pre_virtual_block_data = self._pre_output(self.virtual_block_data)

        # 构建输出数据
        output_data = self._build_output(pre_virtual_block_data)

        # 对输出数据进行后置处理
        result = self._post_output(output_data)

        return result



# -*- coding: utf-8 -*-
from marshmallow import fields

from docparser.config.excel.block_config import BlockConfigSchema


class ExcelDocumentConfigSchema:
    """
    Excel 文档配置
    """
    # 名称
    name = fields.Str()

    # 内容匹配正则
    content_pattern = fields.Str()

    # 文件名匹配正则
    name_pattern = fields.Str()

    """
    采集数据项配置集合
    """
    # 集合
    items = fields.Dict(keys=fields.Str(), values=fields.Nested(BlockConfigSchema))


if __name__ == '__main__':
    cma_config = {
        "type": "table",
        "parser": "mixed",
        "max_rows": 1,
        "row_split_ref_col_name": "container_no",
        "col_split_chars": "  ",
        "rect": {
            "top": {
                "keyword": "CONTAINER  # ",
                "include": True
            },
            "bottom": {
                "keyword": "PLEASE NOTE :",
            }
        },
        "columns": [
            {
                "name": "container_no",
                "title": "CONTAINER #",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\w{0,20}",
            }, {
                "name": "seal_no",
                "title": "SEAL #",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\w{0,20}",
            }, {
                "name": "container_size_type",
                "title": "SIZE/TYPE #",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\d{1,10}\\s{1,2}\\[a-z|A-Z]{2,5}",
            }, {
                "name": "weight",
                "title": "WEIGHT",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\d{0,10}",
            }, {
                "name": "measure",
                "title": "MEASURE",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\w{0,5}",
            }, {
                "name": "free_business_last_free",
                "title": "FREE BUSINESS LAST FREE",
                "title_h_align": "center",
                "title_v_align": "middle",
                "childrens": [
                    {
                        "name": "day_at_port",
                        "title": "DAYS AT PORT",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\w{0,20}",
                    },
                    {
                        "name": "day_at_ramp",
                        "title": "DAY AT RAMP",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\d{1,2}/\\d{1,2}/\\d{1,2}",
                    }
                ]
            }, {
                "name": "pickup_no",
                "title": "PICKUP #",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\w{0,20}",
            },
        ]
    }
    # schema = TableBlockConfigSchema()
    # result = schema.load(cma_config)
    # print(result)



# -*- coding: utf-8 -*-
from marshmallow import Schema, fields
from docparser.config.excel.custom_fields import CustomFields


class RuleConfigSchema(Schema):
    """
    提取规则配置
    """
    # 定位规则
    position_pattern = CustomFields.ActionKeyField(required=True)
    # 单元格查找方向
    find_mode = CustomFields.FindModeField(required=True, default=0)
    # 匹配模式
    separator_mode = CustomFields.SeparatorModeField(required=True, default=0)



# -*- coding: utf-8 -*-

from marshmallow import fields, ValidationError

from docparser.config.excel.enums import BlockType


class CustomFields:
    """
    自定义字段类型
    """

    class BlockTypeField(fields.Field):
        """
        块类型枚举字段
        """

        def _serialize(self, value, attr, obj, **kwargs):
            try:
                if value is None:
                    return BlockType.text.name
                return value.name

            except ValueError as error:
                raise ValidationError("指定的类型值不在[text,table]范围内!") from error

        def _deserialize(self, value, attr, data, **kwargs):
            try:
                return BlockType[value]
            except ValueError as error:
                raise ValidationError("指定的类型值不在[0-1]范围内!") from error

    class TableModeField(fields.Field):
        """
        表格模式
        """

        def _serialize(self, value, attr, obj, **kwargs):
            if value is None:
                return TableMode.standard.name
            return value.name

        def _deserialize(self, value, attr, data, **kwargs):
            try:
                return TableMode[value]
            except ValueError as error:
                raise ValidationError("指定的类型值不在[standard,promiscuous,custom]范围内!") from error

    class HorizontalAlignModeField(fields.Field):
        """
        水平对齐方式
        """

        def _serialize(self, value, attr, obj, **kwargs):
            if value is None:
                return HorizontalAlignMode.right.name
            return value.name

        def _deserialize(self, value, attr, data, **kwargs):
            try:
                return HorizontalAlignMode[value]
            except ValueError as error:
                raise ValidationError("指定的类型值不在[left,center,right]范围内!") from error

    class VerticalAlignModeField(fields.Field):
        """
        水平对齐方式
        """

        def _serialize(self, value, attr, obj, **kwargs):
            if value is None:
                return VerticalAlignMode.middle.name
            return value.name

        def _deserialize(self, value, attr, data, **kwargs):
            try:
                return VerticalAlignMode[value]
            except ValueError as error:
                raise ValidationError("指定的类型值不在[top,middle,bottom]范围内!") from error

    class AnalyticalModeField(fields.Field):
        """
        解析模式枚举字段
        """

        def _serialize(self, value, attr, obj, **kwargs):
            try:
                if value is None:
                    return enums.AnalyticalMode.key_value
                return value.name
            except ValueError as error:
                raise ValidationError("<value:%s 指定的类型值不在[key_value,table]范围内.>" % value) from error

        def _deserialize(self, value, attr, data, **kwargs, ):
            try:
                return enums.AnalyticalMode[value]
            except ValueError as error:
                raise ValidationError("<value:%s 指定的类型值不在[key_value,table]范围内.>" % value) from error

    class FindModeField(fields.Field):
        """
        查找模式枚举字段
        """

        def _serialize(self, value, attr, obj, **kwargs):
            try:
                if value is None:
                    return enums.FindMode.default
                return value.name
            except ValueError as error:
                raise ValidationError("<value:[%s] 指定的类型值不在[default,vertical,horizontal]范围内.>" % value) from error

        def _deserialize(self, value, attr, data, **kwargs, ):
            try:
                return enums.FindMode[value]
            except ValueError as error:
                raise ValidationError("<value:[%s] 指定的类型值不在[default,vertical,horizontal]范围内.>" % value) from error

    class SeparatorModeField(fields.Field):
        """
        分割模式枚举字段
        """

        def _serialize(self, value, attr, obj, **kwargs):
            try:
                if value is None:
                    return enums.SeparatorMode.split
                return value.name
            except ValueError as error:
                raise ValidationError("<value:[%s] 指定的类型值不在[split,regex]范围内.>" % value) from error

        def _deserialize(self, value, attr, data, **kwargs, ):
            try:
                return enums.SeparatorMode[value]
            except ValueError as error:
                raise ValidationError("<value:[%s] 指定的类型值不在[split,regex]范围内.>" % value) from error

    class OverActionField(fields.Field):
        """
        匹配成功后的行为模式枚举字段
        """

        def _serialize(self, value, attr, obj, **kwargs):
            try:
                if value is None:
                    return enums.OverAction.default
                return value.name
            except ValueError as error:
                raise ValidationError("<value:[%s] 指定的类型值不在[default,skip,end]范围内.>" % value) from error

        def _deserialize(self, value, attr, data, **kwargs, ):
            try:
                return enums.OverAction[value]
            except ValueError as error:
                raise ValidationError("<value:[%s] 指定的类型值不在[default,skip,end]范围内.>" % value) from error

    class ActionTypeField(fields.Field):
        """
        赋值模式枚举字段
        """

        def _serialize(self, value, attr, obj, **kwargs):
            try:
                if value is None:
                    return enums.ActionType.add
                return value.name
            except ValueError as error:
                raise ValidationError("<value:[%s] 指定的类型值不在[default,append,cut,split]范围内.>" % value) from error

        def _deserialize(self, value, attr, data, **kwargs, ):
            try:
                return enums.ActionType[value]
            except ValueError as error:
                raise ValidationError("<value:[%s] 指定的类型值不在[default,append,cut,split]范围内.>" % value) from error

    class ActionKeyField(fields.Field):
        """
        数值型或者字符串型
        """

        def _serialize(self, value, attr, obj, **kwargs):
            try:
                if isinstance(value.name, int) or isinstance(value.name, str):
                    return value
                else:
                    raise ValidationError
            except ValueError as error:
                raise ValidationError("<value:[%s] 指定的类型值不在[int,str]范围内.>" % value) from error

        def _deserialize(self, value, attr, data, **kwargs, ):
            try:
                if isinstance(value, int) or isinstance(value, str):
                    return value
                else:
                    raise ValidationError
            except ValueError as error:
                raise ValidationError("<value:[%s] 指定的类型值不在[default,skip,end]范围内.>" % value) from error



# Press the green button in the gutter to run the script.
import os
import pytest



class TestExcelDocumentParser:
    """
    测试Excel文档解析器
    """
    def test_text_block_parse(self):
        """
        测试文本块解析
        """

        converter = DocConverterFactory.create('pdf', 'txt')
        output_file = converter.convert(os.getcwd() + "\\files\\zim.pdf",os.getcwd().replace("\\tests","") +"\\output")

        assert output_file is not None

    def test_standard_table_block_parse(self):
        """
        测试标准表格解析
        """

        converter = DocConverterFactory.create('pdf', 'txt')
        output_files = converter.bulk_convert([os.getcwd() + "\\files\\zim.pdf"], os.getcwd().replace("\\tests","") +r"\output")

        assert len(output_files) == 1

    def test_mixed_table_block_parse(self):
        """
        测试混杂模式表格解析
        """

        converter = DocConverterFactory.create('pdf', 'txt')
        output_files = converter.bulk_convert([os.getcwd() + "\\files\\zim.pdf"],
                                              os.getcwd().replace("\\tests", "") + r"\output")

        assert len(output_files) == 1


if __name__ == '__main__':
    pytest.main("-q --html=report.html")



# -*- coding: utf-8 -*-
from docparser.config.excel.table_block_config import TableBlockConfigSchema


class ConfigManager:
    """
    配置管理器
    """

    @staticmethod
    def load(config):
        schema = TableBlockConfigSchema()
        result = schema.load(config)
        return result








# -*- coding: utf-8 -*-

from marshmallow import Schema, fields


class PositionConfigSchema(Schema):
    """
    定位配置
    """
    # 关键字
    keyword = fields.Str()

    # 水平对齐方式
    h_align = fields.Str()

    # 垂直对齐方式
    v_align = fields.Str()

    # 第几次匹配才满足
    match_times = fields.Integer(default=1)

    # 关键字是否在区间内
    is_outer = fields.Boolean(default=True)


class RectConfigSchema(Schema):
    """
    区域信息
    """
    # 左边关键字位置信息
    left = fields.Nested(PositionConfigSchema())

    # 顶部关键字位置信息
    top = fields.Nested(PositionConfigSchema())

    # 右边关键字位置信息
    right = fields.Nested(PositionConfigSchema())

    # 底部关键字位置信息
    bottom = fields.Nested(PositionConfigSchema())





import time

_easy_cache = {}


def _init():
    global _easy_cache
    _easy_cache = {}


def easy_cache_set(key, value):
    if key not in _easy_cache:
        _easy_cache[key] = {"t": time.time()}
    _easy_cache[key]["value"] = value


def easy_cache_get(key, default=None):
    item = _easy_cache.get(key)
    if not item:
        return default
    if time.time() - item.get("t") > 60 * 60 * 24:
        return default

    return item.get("value")



# -*- coding: utf-8 -*-

from enum import Enum


class BlockType(Enum):
    """
    Excel块类型
    """
    # 文本块
    text = 0

    # 表格块
    table = 1


class TableMode(Enum):
    """
    Excel块表格模式
    """
    # 标准
    standard = 0

    # 混杂模式
    mixed = 1,

    # 自定义
    custom = 2


class HorizontalAlignMode(Enum):
    """
    水平对齐方式
    """
    # 左
    left = 0

    # 中
    center = 1

    # 右
    right = 2


class VerticalAlignMode(Enum):
    """
    垂直对齐方式
    """
    # 上
    top = 0

    # 中
    middle = 1

    # 下
    bottom = 2

class AnalyticalMode(Enum):
    """
    analyticalMode
    解析模式
    1. 键值模式,默认值; 2. 表格模式;
    """
    # 键值模式
    key_value = 1
    # 表格模式
    table = 2


class FindMode(Enum):
    """
    findMode
    查找模式
    1. 纵列模式; 2. 横列模式; 0: 默认模式,为横列模式,且值在本行/本列不向下查找;
    """
    # 默认模式
    default = 0
    # 纵列模式
    v = 1
    # 横列模式
    h = 2


class SeparatorMode(Enum):
    """
    分割模式
    1. 分割字符串成列表，默认值; 2.正则表达式捕获组;
    """
    # 分割字符串成列表
    split = 1
    # 正则表达式捕获组
    regex = 2


class OverAction(Enum):
    """
    匹配成功后的行为
    1: 跳过; 2: 结束; 0:不跳过,默认值;
    """
    # 不跳过,默认值
    row = 0
    # 匹配成功后跳过
    skip = 1
    # 匹配成功后结束
    end = 2


class ActionType(Enum):
    """
    赋值模式
    1. 追加; 0: 覆盖添加,默认值; 2：裁剪模式;  3.分割模式；
    """
    #  覆盖添加,默认值
    add = 0
    # 追加
    append = 1

    # 裁剪模式,仅表格模式下生效,非表格模式下为执行覆盖添加行为
    cut = 2
    # 分割模式,仅表格模式下生效,非表格模式下为执行覆盖添加行为
    split = 3




import abc
import docparser.core.easy_cache
from docparser.core import easy_cache


class BehaviorBase:
    """
    数据处理行为基类
    """
    class_index = 1
    _cache = easy_cache

    def __init__(self):
        pass

    @abc.abstractmethod
    def data_processing(self, ref_data, data: list, error: list, config: dict, logger, additional) -> dict:
        """
        数据处理行为
        :param ref_data: 源数据
        :param data: execl解析出数据
        :param error: execl解析出现的错误
        :param config: 解析配置
        :param logger: 日志工具
        :param additional: 附加数据
        """

    @classmethod
    def _find_values(cls, key, table_name, table):

        values = []
        if table_name and table_name.strip() != "":
            datatable = table.get(table_name)
            for i, x in enumerate(datatable["column"]):
                if x.lower() == key:
                    for j, row in enumerate(datatable["rows"]):
                        values.append(row[i])

        else:
            values.append(table.get(key))
        return values

    @classmethod
    def _find_col_index(cls, key, table_name, table):
        datatable = table.get(table_name)
        for i, x in enumerate(datatable["column"]):
            if x.lower() == key:
                return i


    @classmethod
    def _restore_values(cls, key, table_name, table, values):
        if len(values) == 0:
            return
        if table_name and table_name.strip() != "":
            datatable = table.get(table_name)
            val_len = len(values)
            table_len = len(datatable["rows"])
            for i, x in enumerate(datatable["column"]):
                if x.lower() == key:
                    for j, row in enumerate(datatable["rows"]):
                        if val_len >= table_len:
                            datatable["rows"][j][i] = values[j]
        else:
            table[key] = values[0]



# -*- coding: utf-8 -*-
from docparser.core.text_virtual_block import TextVirtualBlock


class DefaultTextVirtualBlock(TextVirtualBlock):
    """
    默认文本虚拟区块实现
    """

    def __init__(self, sheet, block_config):
        self.virtual_table = None





class ExcelConfig:
    one_config = {
        "id": "group_A",
        "name": "多页模板",
        "kv": {
            "Arrival Vessel ": {
                "position_pattern": [r"^Arrival Vessel :"],
                "value_pattern": [
                    r"Arrival\s*Vessel\s*:\s*(?P<Arrival>[^\n]*?)\nB/L\s*No\s*:\s*(?P<BLNO>[\w]*)(?:\s*|)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Arrival Vessel", "key": 0, "action_type": "add", "default_value": ""},
                    {"keyword": "B/L No", "key": 1, "action_type": "add", "default_value": ""}
                ]
            },
            "Port of Discharging": {
                "position_pattern": [r"^Port of Discharging"],
                "value_pattern": [
                    r"[\w\W]*?(?:\w|\n)(?P<ETA>\d{2}\s*[a-zA-Z]{3,}\s*\d{2}\s*\d{2}\:\d{2}\([a-zA-Z]{2,}\))(?:\s*|\w|\n).*(?P<Available>\d{2}\s*[a-zA-Z]{3,}\s*\d{2}\s*\d{2}\:\d{2})"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "ETA", "key": "ETA", "action_type": "add", "default_value": ""},
                    {"keyword": "Available Date", "key": "Available", "action_type": "add", "default_value": ""}
                ]
            },
            "Est. General Order": {
                "position_pattern": [r"^Est. General Order"],
                "value_pattern": [
                    r"(.*)"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Est. General Order"}
                ]
            },
            "CONTAINER#": {
                "position_pattern": [r"^CONTAINER#"],
                "value_pattern": [
                    r"([a-zA-Z]{4,}\d{7,})"],
                "repeat_count": 1,
                "find_mode": "v",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "CONTAINER#"}
                ]
            },
        }
    }

    hpl_config = {
        "id": "group_A",
        "name": "多页模板",
        "kv": {
            "Ocean Vessel": {
                "position_pattern": [r"^Ocean Vessel:"],
                "value_pattern": [
                    r"(.*)"],
                "repeat_count": 1,
                "find_mode": "v",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Ocean Vessel"}
                ]
            },
            "Voyage No.": {
                "position_pattern": [r"^Voyage No"],
                "value_pattern": [
                    r"^Voyage No\.\s*:\s*(.*)(?:\s*|)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Voyage No"}
                ]
            },
            "Due to arrive at Terminal:": {
                "position_pattern": [r"^Due\s*to\s*arrive\s*at\s*Terminal\s*"],
                "value_pattern": [
                    r"^Due\s*to\s*arrive\s*at\s*Terminal\s*:\s*\n(.*)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Due to arrive at Terminal"}
                ]
            },
            "HBL NO": {
                "position_pattern": [r"[\w\W]*?HBL\s*NO\s*:\s*"],
                "value_pattern": [
                    r"[\w\W]*?HBL\s*NO\s*:\s*(\w*)(?:\n|)"],
                "repeat_count": 1,
                "find_mode": "v",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "HBL NO"}
                ]
            },
        }
    }

    zim_config = {
        "id": "group_A",
        "name": "多页模板",
        "kv": {
            "Vessel/Voyage": {
                "position_pattern": [r"^Vessel/Voyage:"],
                "value_pattern": [
                    r"(.*)"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Vessel/Voyage"}
                ]
            },
            "ETA": {
                "position_pattern": [r"^ETA:"],
                "value_pattern": [
                    r"\n"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "split",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "ETA"}
                ]
            },
            "Date of Loading at": {
                "position_pattern": [r"^Date of Loading at"],
                "value_pattern": [
                    r"(.*)"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Date of Loading at Original Port of Loading"}
                ]
            },
            "Bill of Lading": {
                "position_pattern": [r"^Bill of Lading"],
                "value_pattern": [
                    r"(.*)"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Bill of Lading"}
                ]
            },
            "Container": {
                "position_pattern": [r"^Container"],
                "value_pattern": [
                    r"(.*)\n"],
                "repeat_count": 1,
                "find_mode": "v",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Container"}
                ]
            },
        }
    }

    cmacgm_config = {
        "id": "group_A",
        "name": "多页模板",
        "kv": {
            "VESSEL": {
                "position_pattern": [r"^VESSEL:"],
                "value_pattern": [
                    r"\n"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "split",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "VESSEL"}
                ]
            },
            "VOYAGE": {
                "position_pattern": [r"^\s*VOYAGE\s*:"],
                "value_pattern": [r"^VOYAGE\s*:\s*(?P<VOYAGE>[\w\W]*)$"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "VOYAGE", "key": "VOYAGE", "action_type": "add", "default_value": ""}
                ]
            },
            "LOAD PICKUP POOL ADDRESS": {
                "position_pattern": [r"^LOAD PICKUP POOL ADDRESS:"],
                "value_pattern": [r"(.*)"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "LOAD PICKUP POOL ADDRESS"}
                ]
            },
            "EMPTY RETURN DEPOT": {
                "position_pattern": [r"^EMPTY RETURN DEPOT:"],
                "value_pattern": [r"(.*)"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "EMPTY RETURN DEPOT"}
                ]
            },
            "B/L": {
                "position_pattern": [r"^SCAC     B/L #"],
                "value_pattern": [r".*\s+(\w+)"],
                "repeat_count": 1,
                "find_mode": "v",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "B/L"}
                ]
            },
            "CONTAINER  #": {
                "position_pattern": [r"^CONTAINER  #"],
                "value_pattern": [r"^([a-zA-Z]{4,}\d{4,})"],
                "repeat_count": 1,
                "find_mode": "v",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "CONTAINER"}
                ]
            },
        }
    }

    matson_config = {
        "id": "group_A",
        "name": "多页模板",
        "kv": {
            "Vessel/Voyage": {
                "position_pattern": [r"^Vessel/Voyage:"],
                "value_pattern": [
                    r"\n"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "split",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Vessel/Voyage"}
                ]
            },
            "ETA into: LONG BEACH": {
                "position_pattern": [r"^ETA into: LONG BEACH"],
                "value_pattern": [r"\n"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "split",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "ETA into: LONG BEACH"}
                ]
            },
            "Bill of Lading Number": {
                "position_pattern": [r"^Bill of Lading Number:"],
                "value_pattern": [r"^Bill\s*of\s*Lading\s*Number:\s*([a-zA-Z]{4,}\d{4,})\s*Service"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Bill of Lading Number"}
                ]
            }
        },
        "table": {
            "bill": {
                "position_pattern": [r"^Container#"],
                "separator": "\n",
                "find_mode": "v",
                "separator_mode": "regex",
                "column": ["CONTAINER  #"],
                "behaviors": [
                    {
                        "over_action": "row",
                        "value_pattern": [
                            r"(?P<col_1>[a-zA-Z]{4,}\d{4,}\s*\d+)"],
                        "action": []
                    },
                    {
                        "over_action": "end",
                        "value_pattern": ["^(Total Piece Count)"]
                    }
                ]
            }
        }
    }

    cosco_config = {
        "id": "group_A",
        "name": "多页模板",
        "kv": {
            "B/L Number": {
                "position_pattern": [r"^Type of B/L"],
                "value_pattern": [
                    r"[\w\W]*B/L Number\s*:\s*(?P<Number>(\w+)|())(?:\n|)B/L Vessel/Voyage:(?:\n|)(?P<Vessel>[^\n]*)(?:\n|)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "B/L Number"},
                    {"keyword": "B/L Vessel/Voyage"}
                ]
            },
            "ESTIMATE ARRIVAL AT POD": {
                "position_pattern": [r"[\w\W]*ESTIMATE ARRIVAL AT POD:", r"[\w\W]*?EST CARGO AVAILABLE AT\s*:"],
                "value_pattern": [r"[\w\W]*ESTIMATE ARRIVAL AT POD\s*:\s*(?P<pod_country>\w+)",
                                  r"(?P<POD_time>([\w\W]*?)|(\s|\n|))EST CARGO AVAILABLE AT\s*:\s*(?P<AT_country>[^\n]*?)\n(?P<AT_time>[\w\W]*)(?:\n||$)"],
                "repeat_count": 2,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "ESTIMATE ARRIVAL AT POD Country", "key": "pod_country"},
                    {"keyword": "ESTIMATE ARRIVAL AT POD Time", "key": "POD_time"},
                    {"keyword": "ESTIMATE ARRIVAL AT Country", "key": "AT_country"},
                    {"keyword": "ESTIMATE ARRIVAL AT Time", "key": "AT_time"},
                ]
            }
        },
        "table": {
            "bill": {
                "position_pattern": [r"^CONTAINER NO\.\n\/ SEAL NO\."],
                "separator": "\n",
                "repl_separator": " ",
                "find_mode": "v",
                "separator_mode": "regex",
                "column": ["CONTAINER NO./ SEAL NO."],
                "behaviors": [
                    {
                        "over_action": "row",
                        "value_format": r"([a-zA-Z]{4,}\d{7,}\s*\d{8,})",
                        "value_pattern": [
                            r"(?P<col_1>^[a-zA-Z]{4,}\d{7,}[\w\W]*?\d{7,}$)"],
                        "action": []
                    },
                    {
                        "over_action": "end",
                        "value_pattern": ["^(Remarks)"]
                    }
                ]
            }
        },
        "data_type_format": {
                     "ESTIMATE ARRIVAL AT POD Country": {"table": "", "data_type": "time", "format": "%A, %d %b, %Y %I:%M %p", "filter": "(\\n)"},
                     "ESTIMATE ARRIVAL AT POD Time": {"table": "", "data_type": "time", "format": "%A, %d %b, %Y %I:%M %p", "filter": "(\\n)"},
        }
    }

    evergreen_config = {
        "id": "group_A",
        "name": "多页模板",
        "multi_page": 1,
        "page_pattern": [r"^(ITY OCEAN INTERNATIONAL INC\.,[\w\W]*?CARGO PICK UP LOCATION)(?:\n|$|)"],
        "kv": {
            "ARRIVING VESSEL": {
                "position_pattern": [r"^ARRIVING VESSEL"],
                "value_pattern": [
                    r"ARRIVING VESSEL / VOYAGE NO\.\s*([\w\W]*)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "ARRIVING VESSEL / VOYAGE NO."},
                ]
            },
            "VESSEL ETA": {
                "position_pattern": [r"^VESSEL ETA"],
                "value_pattern": [r"^VESSEL ETA\.\s*([\w\W]*)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "VESSEL ETA"}
                ]
            },
            "CONTAINER NUMBER": {
                "position_pattern": [r"^CONTAINER NUMBER"],
                "value_pattern": [r"CONTAINER NUMBER / CARGO DESCRIPTION\s*([a-zA-Z]{4,}\d{7,}\s*[a-zA-Z]{4,}\d{4,})\s{1,}"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "CONTAINER NUMBER / CARGO DESCRIPTION"}
                ]
            },
            "MASTER’s House B/L No.": {
                "position_pattern": [r"^MASTER’s House B/L No"],
                "value_pattern": [r"MASTER’s House B/L No\.\s*:(?P<no>[\w\W]*)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "MASTER’s House B/L No"}
                ]
            }
        }

    }

    smline_config1 = {
        "id": "group_A",
        "name": "多页模板",
        "kv": {
            "Arrival Vessel": {
                "position_pattern": [r"^Arrival Vessel"],
                "value_pattern": [
                    r"Arrival Vessel\s*:\s*(\w*\s*\w*\s*\d+\s+\w+)\s{4,}[\w\W]*?\nB/L No\s*:\s*([\w]*?)\s{1,}"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Arrival Vessel"},
                    {"keyword": "B/L No"}
                ]
            },
            "ETA": {
                "position_pattern": [r"^ETA/ETB"],
                "value_pattern": [
                    r"[\w\W]*?(\d{2,}\s*[a-zA-Z]{3,}\s*\d{2,}\s*\d{2,}\:\d{2,}\([a-zA-Z]{2}\))[\w\W]*?(\d{2,}\s*[a-zA-Z]{3,}\s*\d{2,}\s*\d{2,}\:\d{2,})[\w\W]*?([a-zA-Z]{3,}\s*\d+\s*[a-zA-Z]{3,})"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "ETA"},
                    {"keyword": "vailable Date"},
                    {"keyword": "Port Free Time"},
                ]
            }
        },
        "table": {
            "bill": {
                "position_pattern": [r"^CONTAINER#"],
                "separator": "\n",
                "find_mode": "h",
                "separator_mode": "regex",
                "column": ["CHG", "RATED AS", "RATE", "PE", "COLLECT"],
                "behaviors": [
                    {
                        "over_action": "row",
                        "loop": 1,
                        "value_pattern": [
                            r"(?P<col_1>[a-zA-Z]*\s{1,}[a-zA-Z]*)\s{1,}(?P<col_2>\d{1,}\.\d{1,})\s{1,}(?P<col_3>\d{1,}\.\d{1,})\s*(?P<col_4>\w{1,})\s{1,}(?P<col_5>\d{1,}\.\d{1,})(?:\n|$)"],
                        "action": []
                    },
                    {
                        "over_action": "end",
                        "value_pattern": ["^(DESTINATION)"]
                    }
                ]
            }
        }
    }

    oocl_config = {
        "id": "group_A",
        "name": "多页模板",
        "kv": {
            "B/L Number": {
                "position_pattern": [r"^Type of B/L"],
                "value_pattern": [
                    r"[\w\W]*B/L Number:\s*([a-zA-Z]+\d+)\s*B/L Vessel/Voyage\s*:\s*(\w+\s+\w+\s+\w+)(?:\n|\s|)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "Vessel/Voyage"},
                    {"keyword": "B/L Number"}

                ]
            },
            "ESTIMATE ARRIVAL AT POD": {
                "position_pattern": [r"^ESTIMATE ARRIVAL AT POD"],
                "value_pattern": [r"ESTIMATE ARRIVAL AT POD:\s*([^\n]*？)\n([\w\W]*)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "ESTIMATE ARRIVAL AT POD1"},
                    {"keyword": "ESTIMATE ARRIVAL AT POD2", "key": 1},
                ]
            },
            "EST CARGO AVAILABLE AT": {
                "position_pattern": [r"^EST CARGO AVAILABLE AT"],
                "value_pattern": [r"^EST CARGO AVAILABLE AT:\s*([^\n]*？)\n([\w\W]*)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "EST CARGO AVAILABLE AT1"},
                    {"keyword": "EST CARGO AVAILABLE AT2", "key": 1}
                ]
            }
        },
        "table": {
            "bill": {
                "position_pattern": [r"^CONTAINER NO\.\n\/ SEAL NO\."],
                "separator": "\n",
                "repl_separator": " ",
                "find_mode": "v",
                "separator_mode": "regex",
                "column": ["CONTAINER NO./ SEAL NO."],
                "behaviors": [
                    {
                        "over_action": "row",
                        "value_format": r"([a-zA-Z]{4,}\d{7,}\s*\d{8,})",
                        "value_pattern": [
                            r"(?P<col_1>^[a-zA-Z]{4,}\d{7,}[\w\W]*?\d{7,}$)"],
                        "action": []
                    },
                    {
                        "over_action": "end",
                        "value_pattern": ["^(Remarks)"]
                    }
                ]
            }
        }
    }

    smline_config = {
        "id": "group_A",
        "name": "多页模板",
        "kv": {
            "Arrival Vessel": {
                "position_pattern": [r"^Arrival Vessel"],
                "value_pattern": [
                    r"Arrival Vessel\s*:\s*(?P<Vessel>\w*\s*\w*\s*\d+\s+\w+)\s{4,}[\w\W]*?\nB/L No\s*:\s*(?P<BillOfLadingsId>[\w]*?)\s{1,}"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "VesselName", "key": "Vessel", "pattern_list": [r"(?P<value>[\w\W]*)\s+?(.{4,})$"]},
                    {"keyword": "VoyageNo", "key": "Vessel", "pattern_list": [r"([\w\W]*)\s+?(?P<value>.{4,})$"]},
                    {"keyword": "BillOfLadingsId", "key": "BillOfLadingsId"}
                ]
            },
            "Port of Discharging": {
                "position_pattern": [r"^Port of Discharging"],
                "value_pattern": [
                    r"\s*Port\s*of\s*Discharging\s(?:\:|\s*)(.*?)(?:\(|$)"],
                "repeat_count": 1,
                "find_mode": "default",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "DestinationPortName"}
                ]
            },
            "ETA": {
                "position_pattern": [r"^ETA/ETB"],
                "value_pattern": [
                    r"\s*(?P<EstimatedArrivalDate>\d{2,}\s*[a-zA-Z]{3,}\s*\d{2,}\s*\d{2,}\:\d{2,})\s*/\s*\d{2,}\s*[a-zA-Z]{3,}\s*\d{2,}\s*\d{2,}\:\d{2,}\s*(?P<DeliveryPlaceName>[a-zA-Z0-9, ]*)(?:\r\n|\n)(?P<DeliveryEstimatedArrivalDate>\d{2,}\s*[a-zA-Z]{3,}\s*\d{2,}\s*\d{2,}\:\d{2,})"],
                "repeat_count": 1,
                "find_mode": "h",
                "separator_mode": "regex",
                "is_split_cell": 0,
                "split_pattern": [""],
                "action": [
                    {"keyword": "EstimatedArrivalDate", "key": "EstimatedArrivalDate"},
                    {"keyword": "DeliveryPlaceName", "key": "DeliveryPlaceName"},
                    {"keyword": "DeliveryEstimatedArrivalDate", "key": "DeliveryEstimatedArrivalDate"},
                ]
            }
        },
        "table": {
            "containers": {
                "position_pattern": [r"^CONTAINER#"],
                "separator": "\n",
                "find_mode": "h",
                "separator_mode": "regex",
                "column": ["ContainerNo"],
                "behaviors": [
                    {
                        "over_action": "row",
                        "loop": 1,
                        "value_pattern": [
                            r"(?P<col_1>[a-zA-Z]{4,}\d{7,})\s*.*?(?:\r\n|\n|$)"],
                        "action": []
                    },
                    {
                        "over_action": "end",
                        "value_pattern": ["^(DESTINATION)"]
                    }
                ]
            }
        },
        "data_type_format": {
            "EstimatedArrivalDate": {"data_type": "time", "format": "%d %b %y %H:%M", "filter": "(\\(CY\\))"},
            "VoyageNo": {"data_type": "str", "filter": " "},
            "DeliveryEstimatedArrivalDate": {"data_type": "time", "format": "%d %b %y %H:%M", "filter": "(\\n)"}
        },
        "address_repair": {
            "db": {
                "pub": {"user": "co", "pwd": "Co&23@2332$22", "server": "db.dev.com:1433",
                        "database": "CO_PUB"}
            },
            "repairs": [
                {"key": "DestinationPortName", "db_key": "pub",
                 "sql": "SELECT  [FullName],[LocalName],[name],[code],[Id] from Places WHERE IsDeleted = 0 and IsOcean = 1 and IsValid = 1 and ([FullName] like '%${value}%' or charindex([FullName],'${value}')> 0) ;",
                 "column": [0, 1, 2, 3], "value": 4, "mapping": "DestinationPortId",
                 "old_val_handle": "empty"},
                {"key": "DeliveryPlaceName", "db_key": "pub",
                 "sql": "SELECT  [FullName],[LocalName],[name],[code],[Id] from Places WHERE IsDeleted = 0 and IsOcean = 1 and IsValid = 1 and ([FullName] like '%${value}%' or charindex([FullName],'${value}')> 0) ;",
                 "column": [0, 1, 2, 3], "value": 4, "mapping": "DeliveryPlaceId",
                 "old_val_handle": "empty"}
            ]

        }
    }

    @staticmethod
    def get_config(name):
        try:
            print("%s_config" % name)
            return getattr(ExcelConfig, "%s_config" % name)
        except:
            return None



# -*- coding: utf-8 -*-
from marshmallow import Schema, fields
from docparser.config.excel.custom_fields import CustomFields


class ActionConfigSchema(Schema):
    """
    匹配行为对象
    """
    # 关键字
    keyword = fields.List(fields.Str(), required=True)

    # 匹配结果的索引
    key = CustomFields.ActionKeyField(required=True)

    # 赋值模式
    action_type = CustomFields.ActionTypeField(required=True)

    # 匹配正则
    pattern_list = fields.List(fields.Str(), default=[])

    # 默认值
    default_value = fields.Str(default=None)



import re
import time
import traceback
from docparser.core.behavior_base import BehaviorBase


class DataTypeBehavior(BehaviorBase):
    """
    数据类型转换处理行为
    =================
    配置说明：主节点下

        "data_type_format": {
                     # table：对应表格模板中的column中的列名，
                     # data_type 对应执行转换方法
                     # format: 格式化的样式
                     # collect: 复杂情况使用正则表达式来处理，必须命名捕获组，且对应format的组名 例如：'{a}+{b}' '(?P<a>\\w*?,)afasdf(?P<a>\\d+)' 'hello,afasdf123123' => 'hello+123123'
                     # filter： 使用正则过滤当前值的干扰字符串
                     # default： 设置默认值的情况下，如果转换失败会使用默认值替换原值
                     "ESTIMATE ARRIVAL AT POD Country": {"table": "", "data_type": "time", "format": "%A, %d %b, %Y %I:%M %p", "filter": "(\\n)"},
                     "ESTIMATE ARRIVAL AT POD Time": {"table": "", "data_type": "time", "format": "%A, %d %b, %Y %I:%M %p", "filter": "(\\n)"},
        }
    """

    class_index = 1

    class TypeConverter:

        @classmethod
        def convert_str(cls, table, error, key, conf, values, collect, val_format, val_filter, default_val, logger):
            return values

        @classmethod
        def convert_time(cls, table, error, key, conf, values, collect, val_format, val_filter, default_val, logger):

            for i in range(len(values)):
                try:
                    values[i] = time.strftime("%Y-%m-%d %H:%M:%S", time.strptime(values[i], val_format))
                except ValueError:
                    if default_val:
                        values[i] = default_val
                    logger.getLogger(cls.__name__).error(
                        f'时间格式转换异常,原因: [{values[i]} | {val_format} | {default_val}] <{traceback.format_exc()}>')
                except Exception:
                    if default_val:
                        values[i] = default_val
                    logger.getLogger(cls.__name__).error(
                        f'时间格式转换异常,原因: [{values[i]} | {val_format} | {default_val}] <{traceback.format_exc()}>')

            return values

    def data_processing(self, ref_data, data: list, error: list, config: dict, logger, additional) -> dict:
        """
        数据处理行为
        :param ref_data: 源数据
        :param data: execl解析出数据
        :param error: execl解析出现的错误
        :param config: 解析配置
        :param logger: 日志工具
        :param additional: 附加数据
        """

        format_config = config.get("data_type_format")
        if data and len(data) > 0 and format_config is not None:
            cls = DataTypeBehavior.TypeConverter()
            for key, conf in format_config.items():

                table_name = conf.get("table")

                data_type = conf.get("data_type")
                fun_name = f"convert_{data_type.lower()}"
                if hasattr(DataTypeBehavior.TypeConverter, fun_name):
                    callback = getattr(cls, fun_name)
                else:
                    continue

                val_format = conf.get("format")
                collect = conf.get("collect")
                val_filter = conf.get("filter")
                default_val = conf.get("default")

                values = self._find_values(key, table_name, data)
                self.__filter(values, val_filter)
                self.__format(values, collect, val_format)
                callback(data, error, key, conf, values, collect, val_format, val_filter, default_val, logger)
                self._restore_values(key, table_name, data, values)

        return additional

    @classmethod
    def __filter(cls, values, rule):
        """
        过滤指定字符
        """
        regex = re.compile(rule)
        for i in range(len(values)):
            values[i] = regex.sub("", values[i])

    @classmethod
    def __format(cls, values, collect, val_format):
        if collect and collect.strip() != "":
            regex = re.compile(collect)
            for i in range(len(values)):
                if (match := regex.search(values[i])) is not None:
                    if (group_dict := match.groupdict()) is not None:
                        values[i] = val_format.format(**group_dict)



# -*- coding: utf-8 -*-
from marshmallow import Schema, fields

from docparser.config.excel.block_config import BlockConfigSchema


class TableColumnConfigSchema(Schema):
    """
    表格列配置架构
    """
    # 名称，未指定名称则不提取数据
    name = fields.Str()

    # 标题
    title = fields.Str(required=True)

    # 数据提取根据改正则模式提取
    title_h_align = fields.Str(required=True, default='center')

    title_v_align = fields.Str(required=True, default='middle')

    # 数据提取根据改正则模式提取
    content_pattern = fields.Str()

    # 数据提取根据改正则模式提取
    content_h_align = fields.Str(default='center')

    content_v_align = fields.Str(default='middle')

    # 子列
    childrens = fields.List(fields.Nested(lambda: TableColumnConfigSchema()))


class TableBlockConfigSchema(BlockConfigSchema):
    """
    表格区块配置架构
    """

    # 最大行数(有些单行情况直接把数据合并到主数据上，并提高采集精度)，
    # -1:不确定,动态多行模式，1:固定一行用于直接行转列
    max_rows = fields.Integer(default=-1)

    # 行分割参考列名
    row_split_ref_col_name = fields.Str()

    # 列分割字符
    col_split_chars = fields.Str()

    # 列定义集合
    columns = fields.List(fields.Nested(TableColumnConfigSchema))


if __name__ == '__main__':
    cma_config = {
        "type": "table",
        "parser": "mixed",
        "max_rows": 1,
        "row_split_ref_col_name": "container_no",
        "col_split_chars": "  ",
        "rect": {
            "top": {
                "keyword": "CONTAINER  # ",
                "include": True
            },
            "bottom": {
                "keyword": "PLEASE NOTE :",
            }
        },
        "columns": [
            {
                "name": "container_no",
                "title": "CONTAINER #",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\w{0,20}",
            }, {
                "name": "seal_no",
                "title": "SEAL #",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\w{0,20}",
            }, {
                "name": "container_size_type",
                "title": "SIZE/TYPE #",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\d{1,10}\\s{1,2}\\[a-z|A-Z]{2,5}",
            }, {
                "name": "weight",
                "title": "WEIGHT",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\d{0,10}",
            }, {
                "name": "measure",
                "title": "MEASURE",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\w{0,5}",
            }, {
                "name": "free_business_last_free",
                "title": "FREE BUSINESS LAST FREE",
                "title_h_align": "center",
                "title_v_align": "middle",
                "childrens": [
                    {
                        "name": "day_at_port",
                        "title": "DAYS AT PORT",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\w{0,20}",
                    },
                    {
                        "name": "day_at_ramp",
                        "title": "DAY AT RAMP",
                        "title_h_align": "center",
                        "title_v_align": "middle",
                        "content_pattern": "\\d{1,2}/\\d{1,2}/\\d{1,2}",
                    }
                ]
            }, {
                "name": "pickup_no",
                "title": "PICKUP #",
                "title_h_align": "center",
                "title_v_align": "middle",
                "content_pattern": "\\w{0,20}",
            },
        ]
    }
    schema = TableBlockConfigSchema()
    result = schema.load(cma_config)
    print(result)



import setuptools
from docparser import __version__

setuptools.setup(
    name='CoDocParser',
    version=__version__,
    packages=setuptools.find_packages("."),  # I also tried exclude=["src/test"]
    url='https://www.cityocean.com',
    license='GPL',
    author='CityOcean',
    author_email='it@cityocean.com',
    description='文档解析器',
    keywords=['document', 'parser'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*",
    extras_require=dict(
        build=['wheel', 'twine']
    ),
    install_requires=[
        'pdfminer', 'marshmallow', 'lxml', 'pandas', 'numpy'
    ],
)



# -*- coding: utf-8 -*-
from marshmallow import fields

from docparser.config.html.html_document_item_config import HtmlDocumentItemConfigSchema


class HtmlDocumentTableColumnConfigSchema(HtmlDocumentItemConfigSchema):
    """
    Html文档文本项配置
    """

    # XPath路径
    xpath = fields.Str()

    # 正则
    pattern = fields.Str()


class HtmlDocumentTableItemConfigSchema(HtmlDocumentItemConfigSchema):
    """
    Html文档表格配置
    """

    # 列
    columns = fields.Dict(keys=fields.Str(), values=fields.Nested(HtmlDocumentTableColumnConfigSchema))




import inspect
import os
import sys

import pandas as pd
from pandas.io.excel import ExcelFile
import numpy as np
import re
import importlib
import importlib.util
from docparser.core.document_parser_base import DocumentParserBase
from docparser.core.kv_data_parser import KvDataParser
from docparser.core.table_data_parser import TableDataParser
from docparser.core.tools import Tools
import logging

logging.basicConfig(format='%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
                    datefmt='%d-%m-%Y:%H:%M:%S',
                    level=logging.ERROR)


class Excel2DocumentParser(DocumentParserBase):
    """
    excel文档解析器
    """

    def __init__(self, file_path, config):
        """
        初始化excel文档解析器
        :param file_path: excel文件位置
        :param config: excel解析匹配
        """
        self.file = file_path
        self.configs = config
        self._table_dict = {}
        self.data = []
        self.errors = []
        self.behaviors = []

        self.__loader_plugins()

        if not os.path.exists(file_path):
            raise FileNotFoundError

    def __read_excel(self, single_page=True):
        """
        读取excel文档
        :return: 原始表数据字典
        """
        data_xls = ExcelFile(self.file)
        # data = {"default": data_xls}
        # self.sheet_names = ["default"]
        data = {}
        self.sheet_names = data_xls.sheet_names
        if single_page:
            data[self.sheet_names[0]] = pd.read_excel(data_xls, sheet_name=self.sheet_names[0], header=None)
            return data

        for name in data_xls.sheet_names:
            df = pd.read_excel(data_xls, sheet_name=name, header=None)
            data[name] = df

        return data

    def __convert_to_table(self, single_page=True):
        """
        转化表单成二维列表
        :return: 处理之后的数据字典
        """
        data_list = self.__read_excel(single_page)
        sheets = {}
        for name, data in data_list.items():
            data = np.array(data).astype("str")
            len_data = len(data)
            execl_list = []
            for row_index in range(len_data):
                row = list(filter(lambda x: not (x == "nan" or x == "NaT"), data[row_index]))
                if len(row) > 0:
                    execl_list.append(row)
            sheets[name] = execl_list
        self._table_dict = sheets

    def __ues_extends(self, sheet, config):
        """
        创建解析器
        :param sheet: 被解析的表
        :param config: 解析配置
        :return: 解析器
        """
        extends_name = os.path.splitext(os.path.split(self.file)[1])[0].split("_")[0]
        _class = None
        if re.match(r'^[a-z][a-zA-Z0-9]*$', extends_name, re.I):
            try:
                module_name = 'docparser.extends.%s_data_parser' % extends_name.lower()
                class_name = '%sDataParser' % extends_name.capitalize()
                data_parser_module = importlib.import_module(module_name)
                cls = getattr(data_parser_module, class_name)
                _class = cls(sheet, config)
            except ModuleNotFoundError:
                pass

        return _class

    def parse(self):
        """
        根据匹配解析数据
        :return: 返回解析后的数据，单个表单页
        """
        self.__convert_to_table()
        return self._extract(self.sheet_names[0], self.configs)

    # def parse_multi_page(self):
    #     """
    #     根据匹配解析数据，存在多个配置时,通用配置，设置键为default,非通用键应该对应excel文档单表名
    #     :return:
    #     """
    #     result = {}
    #     self.__convert_to_table(False)
    #     for name, sheet in self._table_dict.items():
    #         config_name = name if name in self.configs else "default"
    #         result[name] = self._extract(name, self.configs[config_name], config_name)
    #     return result

    def _extract(self, key, config, name="default"):
        """
        合成提取数据字典
        :param key: 需要提取源数据的键
        :param config: 解析配置
        :param config: 配置名称
        :return: 数据字典，错误字典
        """
        multi_page_sheet = [self._table_dict[key]]
        if "multi_page" in self.configs:
            if "page_pattern" not in self.configs:
                self.errors.append(f"<{name}分页配置没有设置分页标识>")
                return self.data, self.errors
            else:
                multi_page_sheet = self._split_sheet(self._table_dict[key])

        for i in range(len(multi_page_sheet)):
            data, errors = self._get_data(multi_page_sheet[i], config, name)
            self.data.append(data)
            self.errors.append(errors)
            if len(errors) > 0:
                break

        return self.data, self.errors

    def _split_sheet(self, sheet):
        """
        分页
        :param sheet: 解析源数据
        :return: 多个数据集
        """
        multi_page_sheet = []
        page_index = -1
        re_page = Tools.init_regex(self.configs["page_pattern"])
        for i in range(len(sheet)):
            row = '\t'.join(sheet[i])
            if Tools.match_find(row, re_page):
                multi_page_sheet.append([])
                page_index += 1
            if page_index > -1:
                multi_page_sheet[page_index].append(sheet[i])
        return multi_page_sheet

    def _get_data(self, sheet, config, name="default"):
        """
        提取数据字典
        :param config: 解析配置
        :param config: 配置名称
        :return: 数据字典，错误字典
        """
        data = {}
        errors = {}
        _class = self.__ues_extends(sheet, config)
        if _class:
            data, errors = _class.parse()
        else:
            if "kv" in config:
                for k, v in config["kv"].items():
                    data_parser = KvDataParser(sheet, data, errors, name)
                    if not data_parser.check_config(v, "kv.%s" % k):
                        return data, errors
                    else:
                        data_parser.parse(v, "kv.%s" % k)
            if "table" in config:
                for k, v in config["table"].items():
                    data_parser = TableDataParser(sheet, data, errors, name)
                    if not data_parser.check_config(v, "table.%s" % k):
                        return data, errors
                    else:
                        data_parser.parse(v, "table.%s" % k)
            additional = {}
            for cls in self.behaviors:
                try:
                    additional = cls().data_processing(sheet, data, errors, config, logging, additional)
                except Exception as ex:
                    if "BehaviorsError" not in errors:
                        errors["BehaviorsError"] = []
                    errors["BehaviorsError"].append(ex.args)

        return data, errors

    def __loader_plugins(self):
        """
        加载解析后的处理行为
        """
        module_name = "docparser.plugins"
        if module_name in sys.modules:
            module = sys.modules["docparser.plugins"]
        elif (spec := importlib.util.find_spec(module_name)) is not None:
            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module
            spec.loader.exec_module(module)
        else:
            return None

        plugins = list(filter(lambda x: x.endswith("_behavior.py"), os.listdir(module.__path__[0])))
        for file in plugins:
            module_file = os.path.splitext(os.path.basename(file))[0]
            module = importlib.import_module(f"{module_name}.{module_file}")
            cls = getattr(module,
                          ''.join([s.capitalize() for s in os.path.splitext(os.path.basename(file))[0].split('_')]))
            if inspect.isclass(cls):
                self.behaviors.append(cls)
        if len(self.behaviors) > 0:
            self.behaviors = sorted(self.behaviors, key=lambda x: x.class_index)



import json
import os
import pytest
from docparser.doc_parser_factory import DocParserFactory

# from tests.excel_config import ExcelConfig as Config
doc1_config = {}
test_config = {
    "id": "CMA",
    "name": "CMA config",
    "kv": {
        "VESSEL": {
            "position_pattern": [
                "^VESSEL:"
            ],
            "value_pattern": [
                "(?P<Vessel>[\\w\\W]*?)(?:\\r\\n|\\n|$)"
            ],
            "repeat_count": 1,
            "find_mode": "h",
            "separator_mode": "regex",
            "is_split_cell": 0,
            "split_pattern": [
                ""
            ],
            "action": [
                {
                    "keyword": "VesselName",
                    "key": "Vessel"
                }
            ]
        },
        "VOYAGE": {
            "position_pattern": [
                "^VOYAGE:"
            ],
            "value_pattern": [
                "VOYAGE\\s*:\\s*(?P<VOYAGE>[\\w\\W]*)"
            ],
            "repeat_count": 1,
            "find_mode": "default",
            "separator_mode": "regex",
            "is_split_cell": 0,
            "split_pattern": [
                ""
            ],
            "action": [
                {
                    "keyword": "VoyageNo",
                    "key": "VOYAGE"
                }
            ]
        },
        "POD ETA": {
            "position_pattern": [
                "^POD ETA"
            ],
            "value_pattern": [
                "POD\\s*ETA\\s*:\\s*(?P<ETA>\\d+/\\d+/\\d+)(?:\\r\\n|\\n)"
            ],
            "repeat_count": 1,
            "find_mode": "default",
            "separator_mode": "regex",
            "is_split_cell": 0,
            "split_pattern": [
                ""
            ],
            "action": [
                {
                    "keyword": "EstimatedArrivalDate",
                    "key": "ETA"
                }
            ]
        },
        "DeliveryPlaceName": {
            "position_pattern": [
                "^OPERATIONAL LOAD PORT"
            ],
            "value_pattern": [
                "[\\w\\W]*?(?:\\n|\\r\\n|)(?P<DELIVERY>.*)"
            ],
            "repeat_count": 1,
            "find_mode": "h",
            "separator_mode": "regex",
            "is_split_cell": 0,
            "split_pattern": [
                ""
            ],
            "action": [
                {
                    "keyword": "DeliveryPlaceName",
                    "key": "DELIVERY"
                }
            ]
        },
        "BillOfLadingsId": {
            "position_pattern": [
                "^Please clear your cargo",
                "Please Pay freight against "
            ],
            "value_pattern": [
                "[\\w\\W]*?(?P<billoflading>[a-zA-Z]{4,}\\d{5,}[a-zA-Z]*)\\s*",
                "[\\w\\W]*?(?P<billoflading>[a-zA-Z]{4}\\s*[a-zA-Z]{3,}\\d{5,}[a-zA-Z]*)\\s*((Waybill)|(Negotiable))"
            ],
            "repeat_count": 1,
            "find_mode": "default",
            "separator_mode": "regex",
            "is_split_cell": 0,
            "split_pattern": [
                ""
            ],
            "action": [
                {
                    "keyword": "BillOfLadingsId",
                    "key": "billoflading"
                }
            ]
        },
        "BillOfLadingsId1": {
            "position_pattern": [
                "^Please[\\w\\W]*BILL\\s*TYPE$",
                "^SCAC\\s{2,}B/L\\s*#$"
            ],
            "value_pattern": [
                "(?P<billoflading>[a-zA-Z]{4,}\\s*[a-zA-Z]*\\d{5,}[a-zA-Z]*)$",
            ],
            "repeat_count": 1,
            "find_mode": "v",
            "separator_mode": "regex",
            "is_split_cell": 0,
            "split_pattern": [
                ""
            ],
            "action": [
                {
                    "keyword": "BillOfLadingsId",
                    "key": "billoflading"
                }
            ]
        }
    },
    "table": {
        "containers": {
            "position_pattern": [
                "^CONTAINER\\s*#"
            ],
            "separator": " ",
            "find_mode": "v",
            "separator_mode": "regex",
            "column": [
                "ContainerNo"
            ],
            "behaviors": [
                {
                    "over_action": "row",
                    "value_pattern": [
                        "(?P<col_1>([a-zA-Z]{4,}\\d{7,}\\s*)*)"
                    ],
                    "action": []
                }
            ]
        }
    },
    "data_type_format": {
        "VoyageNo": {
            "data_type": "str",
            "filter": "r([/\\s])"
        },
        "EstimatedArrivalDate": {
            "data_type": "time",
            "format": "%m/%d/%Y",
            "filter": ""
        },
        "BillOfLadingsId": {
            "data_type": "str",
            "filter": "(\\s)"
        }
    },
    # "address_repair": {
    #     "db": {
    #         "pub": {
    #             "user": "co",
    #             "pwd": "Co&23@2332$22",
    #             "server": "db.dev.com:1433",
    #             "database": "CO_PUB"
    #         }
    #     },
    #     "repairs": [
    #         {
    #             "key": "DeliveryPlaceName",
    #             "db_key": "pub",
    #             "sql": "SELECT  [FullName],[LocalName],[name],[code],[Id] from Places WHERE IsDeleted = 0 and IsOcean = 1 and IsValid = 1 and ([FullName] like '%${value}%' or charindex([FullName],'${value}')> 0) ;",
    #             "column": [
    #                 0,
    #                 1,
    #                 2,
    #                 3
    #             ],
    #             "value": 4,
    #             "mapping": "DeliveryPlaceId",
    #             "old_val_handle": "empty"
    #         }
    #     ]
    # }
}

config_names = []

for k, v in test_config.get('kv').items():
    action = v.get('action')
    for _item in action:
        config_names.append((k, _item.get("keyword")))


class TestExcelDocumentParser:

    def test_excel_file_parse(self):
        """
        单文件测试
        :return:
        """
        name = "cma".upper()

        doc1_config["id"] = f"AN_{name}_"

        test_config["id"] = name
        test_config["name"] = f"{name} config"
        doc1_config["parse"] = test_config
        # print(test_config)
        print(json.dumps(doc1_config))
        str_dir = f"C:\\Users\\APing\\Desktop\\temp\\{name}"
        files = os.listdir(str_dir)
        size = 0
        f_list = []
        for f in files:
            # if f != "bl2687093129.xlsx":
            #     continue
            if f.endswith('.xlsx') and not f.startswith('~'):
                # factory = DocParserFactory.create("excel2",
                #                                   r"C:\Users\APing\Desktop\temp\cma\ARR_JXTL311318.xlsx",
                #                                   test_config)
                factory = DocParserFactory.create("excel2", os.path.join(str_dir, f), test_config)
                result, errors = factory.parse()
                size += 1
                if len(result) == 0:
                    f_list.append({f: errors})
                else:
                    err_list = []
                    for nm in config_names:
                        v1 = result[0].get(nm[1])
                        if (v1 is None or v1 == "") and len(errors[0]) > 0 and len(
                                [x for x in errors[0].keys() if x.find(nm[0]) > -1]) > 0:
                            err_list.append(nm[1])
                            break
                    if len(err_list) > 0:
                        f_list.append({f: (err_list, errors)})

                print(f, result, errors)
        print(f"一共解析文件数量：{size}, err-size: {len(f_list)}")
        for _nm in f_list:
            print(_nm)
    # def test_excel_dir_parse(self):
    #     """
    #     测试文件夹下的拥有对应名称配置的excel文件
    #     :return:
    #     """
    #     path = os.getcwd() + "\\files"
    #     dirs = os.listdir(path)
    #     for file in dirs:
    #         name = file.split(".")[0]
    #         if ".xlsx" in file:
    #             _config = Config.get_config(name.lower())
    #             if _config is None:
    #                 continue
    #             factory = DocParserFactory.create("excel2", "%s\\%s.xlsx" % (path, name.lower()), _config)
    #             result, errors = factory.parse()
    #             print("=========================", file, "========================")
    #             print(_config)
    #             print(path + file)
    #             print(result)
    #             print(errors)
    #             print("------------------------------------------------------------")
    #             print("\r\n\r\n")


if __name__ == '__main__':
    pytest.main("-q --html=report.html")
